"""
The Simulator class is the access point for the main features of NengoDL,
including `running <.Simulator.run_steps>` and `training <.Simulator.train>`
a model.
"""

import collections
import copy
import logging
import os
import tempfile
import warnings

from nengo import Ensemble, Connection, Probe, Network, Direct, Node
from nengo.builder.connection import BuiltConnection
from nengo.builder.ensemble import BuiltEnsemble
from nengo.ensemble import Neurons
from nengo.exceptions import (
    ReadonlyError,
    SimulatorClosed,
    NengoWarning,
    SimulationError,
    ValidationError,
)
from nengo.solvers import NoSolver
import numpy as np
import tensorflow as tf
from tensorflow.python.client.timeline import Timeline
from tensorflow.python.ops import gradient_checker

from nengo_dl import utils, config, objectives
from nengo_dl.builder import NengoBuilder, NengoModel
from nengo_dl.compat import tf_compat, Convolution
from nengo_dl.tensor_graph import TensorGraph


logger = logging.getLogger(__name__)


class Simulator:
    """
    Simulate network using the ``nengo_dl`` backend.

    Parameters
    ----------
    network : `~nengo.Network` or None
        A network object to be built and then simulated. If None,
        then a built model must be passed to ``model`` instead
    dt : float
        Length of a simulator timestep, in seconds
    seed : int
        Seed for all stochastic operators used in this simulator
    model : `~nengo.builder.Model`
        Pre-built model object
    dtype : ``tf.DType``
        Deprecated, use ``nengo_dl.configure_settings(dtype=...)`` instead.
    device : None or ``"/cpu:0"`` or ``"/gpu:[0-n]"``
        Device on which to execute computations (if None then uses the
        default device as determined by TensorFlow)
    unroll_simulation : int
        Unroll simulation loop by explicitly building the given number of
        iterations into the computation graph (improves simulation speed
        but increases build time)
    minibatch_size : int
        The number of simultaneous inputs that will be passed through the
        network
    tensorboard : str
        If not None, save network output in the TensorFlow summary format to
        the given directory, which can be loaded into TensorBoard
    progress_bar : bool
        If True (default), display progress information when building a model
    """

    def __init__(
        self,
        network,
        dt=0.001,
        seed=None,
        model=None,
        dtype=None,
        device=None,
        unroll_simulation=1,
        minibatch_size=None,
        tensorboard=None,
        progress_bar=True,
    ):
        self.closed = None
        self.unroll = unroll_simulation
        self.minibatch_size = 1 if minibatch_size is None else minibatch_size
        self.data = SimulationData(self, minibatch_size is not None)
        if seed is None:
            if network is not None and network.seed is not None:
                seed = network.seed + 1
            else:
                seed = np.random.randint(np.iinfo(np.int32).max)
        self.seed = seed

        # TODO: multi-GPU support

        if device is None and not utils.tf_gpu_installed:
            warnings.warn(
                "No GPU support detected. It is recommended that you "
                "install tensorflow-gpu (`pip install tensorflow-gpu`)."
            )
            logger.info("Running on CPU")
        else:
            logger.info(
                "Running on %s",
                "CPU/GPU" if device is None else ("CPU" if "cpu" in device else "GPU"),
            )

        ProgressBar = utils.ProgressBar if progress_bar else utils.NullProgressBar

        # build model (uses default nengo builder)
        if model is None:
            self.model = NengoModel(
                dt=float(dt),
                label="%s, dt=%f" % (network, dt),
                builder=NengoBuilder(),
                fail_fast=False,
            )
        else:
            if dt != model.dt:
                warnings.warn(
                    "Model dt (%g) does not match Simulator "
                    "dt (%g)" % (model.dt, dt),
                    NengoWarning,
                )
            self.model = model

        if network is not None:
            p = ProgressBar("Building network", "Build")
            self.model.build(network, progress=p)

        if dtype is not None:
            warnings.warn(
                "dtype parameter is deprecated; use "
                "nengo_dl.configure_settings(dtype=...) instead",
                DeprecationWarning,
            )
        else:
            dtype = config.get_setting(self.model, "dtype", tf.float32)

        # set up tensorflow graph plan
        with ProgressBar(
            "Optimizing graph", "Optimization", max_value=None
        ) as progress:
            self.tensor_graph = TensorGraph(
                self.model,
                self.dt,
                unroll_simulation,
                dtype,
                self.minibatch_size,
                device,
                progress,
            )

        # set TensorFlow graph seed
        with self.tensor_graph.graph.as_default():
            tf_compat.set_random_seed(self.seed)

        # construct graph
        with ProgressBar(
            "Constructing graph", "Construction", max_value=None
        ) as progress:
            self.tensor_graph.build(progress)

        # output simulation data for viewing via TensorBoard
        if tensorboard is not None:
            if not os.path.exists(tensorboard):
                os.makedirs(tensorboard)

            run_number = (
                max(
                    [int(x[4:]) for x in os.listdir(tensorboard) if x.startswith("run")]
                    or [-1]
                )
                + 1
            )
            self.summary = tf_compat.summary.FileWriter(
                os.path.join(tensorboard, "run_%d" % run_number),
                graph=self.tensor_graph.graph,
            )
        else:
            self.summary = None

        # start session

        session_config = tf_compat.ConfigProto(
            allow_soft_placement=False, log_device_placement=False
        )

        # TODO: XLA compiling doesn't seem to provide any benefit at the
        # moment, revisit later after tensorflow has developed it further
        # config.graph_options.optimizer_options.global_jit_level = (
        #     tf.OptimizerOptions.ON_1)

        # set any config options specified by user
        config_settings = config.get_setting(self.model, "session_config", {})
        for c, v in config_settings.items():
            attrs = c.split(".")
            x = session_config
            for a in attrs[:-1]:
                x = getattr(x, a)
            setattr(x, attrs[-1], v)

        self.sess = tf_compat.Session(
            graph=self.tensor_graph.graph, config=session_config
        )

        self.reset(seed=seed)

        self.closed = False

    def reset(self, seed=None):
        """
        Resets the simulator to initial conditions.

        Parameters
        ----------
        seed : int
            If not None, overwrite the default simulator seed with this value
            (note: this becomes the new default simulator seed)
        """

        if self.closed is not None and self.closed:
            raise SimulatorClosed("Cannot reset closed Simulator.")

        self.n_steps = 0
        self.time = 0.0

        # initialize variables
        self.sess.run(
            self.tensor_graph.constant_init_op,
            feed_dict=self.tensor_graph.signals.constant_phs,
        )
        self.soft_reset(include_trainable=True, include_probes=True)

        # execute post-build processes (we do this here because
        # seed can change each call to reset)
        if seed is not None:
            self.seed = seed
        self.rng = np.random.RandomState(self.seed)
        with self.tensor_graph.graph.as_default():
            tf_compat.set_random_seed(self.seed)

        with self.sess.as_default():
            self.tensor_graph.build_post(self.sess, self.rng)

    def soft_reset(self, include_trainable=False, include_probes=False):
        """
        Resets the internal state of the simulation, but doesn't
        rebuild the graph.

        Parameters
        ----------
        include_trainable : bool
            If True, also reset any training that has been performed on
            network parameters (e.g., connection weights)
        include_probes : bool
            If True, also clear probe data
        """

        init_ops = [self.tensor_graph.local_init_op, self.tensor_graph.global_init_op]
        if include_trainable:
            init_ops.append(self.tensor_graph.trainable_init_op)
        self.sess.run(
            init_ops,
            feed_dict={ph: v for _, ph, v in self.tensor_graph.base_vars.values()},
        )

        if include_probes:
            for p in self.model.probes:
                self.model.params[p] = []
            self.n_steps = 0

    def step(self, **kwargs):
        """
        Run the simulation for one time step.

        Parameters
        ----------
        kwargs : dict
            See `.run_steps`

        Notes
        -----
        Progress bar is disabled by default when running via this method.
        """

        kwargs.setdefault("progress_bar", False)

        self.run_steps(1, **kwargs)

    def run(self, time_in_seconds, **kwargs):
        """
        Simulate for the given length of time.

        Parameters
        ----------
        time_in_seconds : float
            Run the simulator for the given number of simulated seconds
        kwargs : dict
            See `.run_steps`
        """

        if time_in_seconds < 0:
            raise ValidationError(
                "Must be positive (got %g)" % (time_in_seconds,), attr="time_in_seconds"
            )

        steps = int(np.round(float(time_in_seconds) / self.dt))

        if steps == 0:
            warnings.warn(
                "%g results in running for 0 timesteps. Simulator "
                "still at time %g." % (time_in_seconds, self.time)
            )
        else:
            self.run_steps(steps, **kwargs)

    def run_steps(
        self,
        n_steps,
        data=None,
        input_feeds=None,
        profile=False,
        progress_bar=True,
        extra_feeds=None,
    ):
        """
        Simulate for the given number of steps.

        Parameters
        ----------
        n_steps : int
            The number of simulation steps to be executed
        data : dict of {`~nengo.Node`: `~numpy.ndarray`}
            Override the values of input Nodes with the given data.  Arrays
            should have shape ``(sim.minibatch_size, n_steps, node.size_out)``.
        input_feeds : dict of {`~nengo.Node`: `~numpy.ndarray`}
            Deprecated, use ``data`` instead.
        profile : bool
            If True, collect TensorFlow profiling information while the
            simulation is running (this will slow down the simulation).
            Can also pass a string specifying a non-default filename for the
            saved profile data.
        progress_bar : bool
            If True, print information about the simulation status to standard
            output.
        extra_feeds : dict of {``tf.Tensor``: `~numpy.ndarray`}
            Can be used to feed a value for arbitrary Tensors in the simulation
            (will be passed directly to the TensorFlow session)

        Notes
        -----
        If ``unroll_simulation=x`` is specified, and ``n_steps > x``, this will
        repeatedly execute ``x`` timesteps until the the number of steps
        executed is >= ``n_steps``.
        """

        actual_steps = self.unroll * int(np.ceil(n_steps / self.unroll))

        if actual_steps != n_steps:
            warnings.warn(
                "Number of steps (%d) is not an even multiple of "
                "`unroll_simulation` (%d).  Simulation will run for %d steps, "
                "which may have unintended side effects."
                % (n_steps, self.unroll, actual_steps),
                RuntimeWarning,
            )

        if input_feeds is not None:
            # TODO: remove this in 3.0.0
            warnings.warn(
                "The `input_feeds` argument has been renamed; please use "
                "`data` instead, as `input_feeds` will not be supported in a "
                "future version.",
                DeprecationWarning,
            )
            assert data is None
            data = input_feeds

        if data is None:
            data = actual_steps
        else:
            # note: we only need to check the shape of the first item, because
            # check_data (inside run_batch) will ensure that all the items
            # have the same shape
            batch_size, input_steps = next(iter(data.values())).shape[:2]
            if batch_size != self.minibatch_size:
                raise ValidationError(
                    "Input data must have batch size == sim.minibatch_size "
                    "(%d != %d)" % (batch_size, self.minibatch_size),
                    "data",
                )
            if input_steps != actual_steps:
                raise ValidationError(
                    "Number of timesteps in input data (%d) does not "
                    "match requested number of steps (%d)" % (input_steps, n_steps),
                    "data",
                )

        def callback(_, extra_vals):
            assert extra_vals == actual_steps

        progress = (
            utils.ProgressBar("Simulating", "Simulation", max_value=None)
            if progress_bar
            else utils.NullProgressBar()
        )

        with progress:
            # note: we request steps_run from extra_fetches so that the
            # simulation will always run for the given number of steps, even
            # if there are no output probes
            probe_data = self.run_batch(
                data,
                {p: None for p in self.model.probes},
                extra_feeds=extra_feeds,
                extra_fetches=self.tensor_graph.steps_run,
                combine=lambda x: x[0],
                isolate_state=False,
                callback=callback,
                profile=profile,
            )

        # update stored probe data
        for probe, val in probe_data.items():
            # drop any extra steps (due to uneven unroll_simulation)
            val = val[:, :n_steps]

            if probe.sample_every is not None:
                # downsample probe according to `sample_every`
                period = probe.sample_every / self.dt
                steps = np.arange(self.n_steps, self.n_steps + n_steps)
                val = val[:, (steps + 1) % period < 1]

            self.model.params[probe].append(val)

        # update n_steps
        # note: we update n_steps according to the number of steps that the
        # user asked for, not the number of steps that were actually run
        # (in the case of uneven unroll_simulation)
        self.n_steps += n_steps
        self.time = self.n_steps * self.dt

    def train(
        self,
        data,
        optimizer,
        n_epochs=1,
        objective=None,
        shuffle=True,
        truncation=None,
        summaries=None,
        profile=False,
        extra_feeds=None,
        progress_bar=True,
    ):
        """
        Optimize the trainable parameters of the network using the given
        optimization method, minimizing the objective value over the given
        inputs and targets.

        Parameters
        ----------
        data : dict of {`~nengo.Node` or `~nengo.Probe`: \
                        `~numpy.ndarray`} or int
            Input values for Nodes in the network or target values for Probes;
            arrays should have shape ``(batch_size, n_steps,
            node.size_out/probe.size_in)``.  If no input data is required,
            an integer can be given specifying the number of timesteps to
            run the simulation.
        optimizer : ``tf.train.Optimizer``
            TensorFlow optimizer, e.g.
            ``tf.train.GradientDescentOptimizer(learning_rate=0.1)``
        n_epochs : int
            Run training for the given number of epochs (complete passes
            through ``data``)
        objective : dict of {(tuple of) `~nengo.Probe`: callable or ``None``}
            The objective to be minimized. The default applies
            `.objectives.mse` to all probes in ``data``.  This can be
            overridden by passing a dictionary mapping Probes to functions
            ``f(output, target) -> loss`` that consume the actual output and
            target output for the given probe(s) and return a ``tf.Tensor``
            representing a scalar loss value.  The function may also accept a
            single argument ``f(output) -> loss`` if targets are not required.
            Some common objective functions can be found in
            `nengo_dl.objectives`.

            Passing ``None`` as the probe value (instead of a callable)
            indicates that the error is being computed outside the simulation,
            and the value passed for that probe in ``data`` directly specifies
            the output error gradient.

            If multiple probes are specified as the key, then the corresponding
            output/target values will be passed as a list to the objective
            function.

            The overall loss value being minimized will be the sum across all
            the objectives specified.
        shuffle : bool
            If True, randomize the data into different minibatches each epoch
        truncation: int
            If not None, use truncated backpropagation when training the
            network, with the given truncation length.
        summaries : list of `~nengo.Connection` or \
                            `~nengo.Ensemble` or \
                            `~nengo.ensemble.Neurons` or \
                            ``"loss"`` or \
                            ``tf.Tensor``
            If not None, collect data during the training process using
            TensorFlow's ``tf.summary`` format.  The summary objects can be a
            Connection (in which case data on the corresponding weights will be
            collected), Ensemble (encoders), Neurons (biases), or ``"loss"``
            (the loss value for ``objective``).  The user can also create their
            own summaries and pass in the Tensors representing the summary ops.
        profile : bool
            If True, collect TensorFlow profiling information while the
            simulation is running (this will slow down the simulation).
            Can also pass a string specifying a non-default filename for the
            saved profile data.
        extra_feeds : dict of {``tf.Tensor``: `~numpy.ndarray`}
            Can be used to feed a value for arbitrary Tensors in the simulation
            (will be passed directly to the TensorFlow session)
        progress_bar : bool
            If True, print information about the simulation status to standard
            output.

        Notes
        -----
        Most deep learning methods require the network to be differentiable,
        which means that trying to train a network with non-differentiable
        elements will result in an error.  Examples of common
        non-differentiable elements include `~nengo.LIF`,
        `~nengo.Direct`, or processes/neurons that don't have a
        custom TensorFlow implementation (see
        `.process_builders.SimProcessBuilder`/
        `.neuron_builders.SimNeuronsBuilder`)
        """

        if isinstance(data, int):
            batch_size = self.minibatch_size
            n_steps = data
        else:
            batch_size, n_steps = next(iter(data.values())).shape[:2]

        # error checking
        synapses = [
            x.synapse is not None
            for x in (
                self.model.toplevel.all_connections
                + (
                    list(p for p in data if isinstance(p, Probe))
                    if isinstance(data, dict)
                    else []
                )
            )
        ]
        if n_steps == 1 and self.model.toplevel is not None and any(synapses):
            warnings.warn(
                "Training for one timestep, but the network contains "
                "synaptic filters (which will introduce at least a "
                "one-timestep delay); did you mean to set synapse=None?"
            )
        if isinstance(optimizer, dict):
            raise ValidationError(
                "The second argument to `sim.train` should be a "
                "tf.train.Optimizer, not a dictionary; it is likely that this "
                "code was written for NengoDL 1.x and needs to be updated for "
                "NengoDL 2.x; see "
                "https://www.nengo.ai/nengo-dl/project.html#release-history",
                "optimizer",
            )

        # fill in default objective
        if objective is None:
            if isinstance(data, int):
                raise ValidationError(
                    "Must specify an explicit objective if no input data given",
                    "objective",
                )
            objective = {p: objectives.mse for p in data if isinstance(p, Probe)}

        if not isinstance(objective, dict):
            raise ValidationError(
                "Must be a dictionary mapping Probes to objective functions",
                "objective",
            )

        # fill in mse function
        for p, o in objective.items():
            if o == "mse":
                # TODO: remove in 3.0.0
                warnings.warn(
                    "Using the string 'mse' for the objective is deprecated, "
                    "and will no longer be supported in the future; please "
                    "use the function `nengo_dl.objectives.mse` in the future",
                    DeprecationWarning,
                )

                objective[p] = objectives.mse

        # build the output function
        apply_optimizer = self.tensor_graph.build_optimizer_func(optimizer, objective)

        extra_fetches = dict()

        # add summaries
        if summaries is not None:
            if self.summary is None:
                warnings.warn(
                    "Simulator was created with tensorboard=False; "
                    "ignoring requested summaries"
                )
            else:
                for i, v in enumerate(summaries):
                    if isinstance(v, str) and v == "loss":
                        summaries[i] = objective
                summary_op, init = self.tensor_graph.build_summaries(summaries)
                if init is not None:
                    # initialize any variables created when building summaries
                    self.sess.run(init)
                extra_fetches["summaries"] = summary_op

        progress = (
            utils.ProgressBar(
                "Training",
                max_value=(
                    n_epochs
                    * (batch_size // self.minibatch_size)
                    * (1 if truncation is None else n_steps // truncation)
                ),
                vars=["loss"],
            )
            if progress_bar
            else utils.NullProgressBar()
        )

        objective_probes = tuple(objective.keys())

        def callback(out_vals, extra_vals):
            # update progress bar, with loss value
            loss = out_vals[objective_probes][1]
            # loss will be {} if only direct grads used when calculating
            # gradient
            kwargs = {} if loss == {} else dict(loss="%.4f" % loss)
            progress.step(**kwargs)

            # export summaries to tensorboard
            if "summaries" in extra_vals:
                # note: the first output value is the new value of the
                # global training_step
                self.summary.add_summary(
                    extra_vals["summaries"], out_vals[objective_probes][0]
                )

        # run training
        with progress:
            self.run_batch(
                data,
                {objective_probes: apply_optimizer},
                n_epochs=n_epochs,
                combine=lambda x: None,
                extra_feeds=extra_feeds,
                extra_fetches=extra_fetches,
                truncation=truncation,
                profile=profile,
                shuffle=shuffle,
                training=True,
                callback=callback,
            )

    def loss(
        self,
        data,
        objective=None,
        combine=np.mean,
        extra_feeds=None,
        progress_bar=True,
        training=False,
    ):
        """
        Compute the loss value for the given objective and inputs/targets.

        Parameters
        ----------
        data : dict of {`~nengo.Node` or `~nengo.Probe`: \
                        `~numpy.ndarray`} or int
            Input values for Nodes in the network or target values for Probes;
            arrays should have shape ``(batch_size, n_steps,
            node.size_out/probe.size_in)``.  If no input data is required,
            an integer can be given specifying the number of timesteps to
            run the simulation.
        objective : dict of {(tuple of) `~nengo.Probe`: callable}
            The objective to compute the loss. The default applies
            `.objectives.mse` to all probes in ``data``.  This can be
            overridden by passing a dictionary mapping Probes to functions
            ``f(output, target) -> loss`` that consume the actual output and
            target output for the given probe(s) and return a ``tf.Tensor``
            representing a scalar loss value.  The function may also accept a
            single argument ``f(output) -> loss`` if targets are not required.
            Some common objective functions can be found in
            `nengo_dl.objectives`.

            If multiple probes are specified as the key, then the corresponding
            output/target values will be passed as a list to the objective
            function.

            The overall value returned will be the sum across all
            the objectives specified.
        combine : callable
            Function used to combine objective values from each minibatch.
        extra_feeds : dict of {``tf.Tensor``: `~numpy.ndarray`}
            Can be used to feed a value for arbitrary Tensors in the simulation
            (will be passed directly to the TensorFlow session)
        progress_bar : bool
            If True, print information about the simulation status to standard
            output.
        training : bool
            If True, run the network in training mode (where, e.g., spiking
            neuron models are swapped for the equivalent differentiable
            approximation).

        Returns
        -------
        loss : float
            Sum of computed error values for each function in ``objective``.
        """

        batch_size = (
            self.minibatch_size
            if isinstance(data, int)
            else next(iter(data.values())).shape[0]
        )

        # fill in default objective
        if objective is None:
            if isinstance(data, int):
                raise ValidationError(
                    "Must specify an explicit objective if no input data given",
                    "objective",
                )
            objective = {p: objectives.mse for p in data if isinstance(p, Probe)}

        if not isinstance(objective, dict):
            raise ValidationError(
                "Must be a dictionary mapping Probes to objective functions",
                "objective",
            )

        # fill in mse function
        for p, o in objective.items():
            if o == "mse":
                # TODO: remove in 3.0.0
                warnings.warn(
                    "Using the string 'mse' for the objective is deprecated, "
                    "and will no longer be supported in the future; please "
                    "use the function `nengo_dl.objectives.mse` in the future",
                    DeprecationWarning,
                )

                objective[p] = objectives.mse

        progress = (
            utils.ProgressBar(
                "Calculating loss",
                "Calculation",
                max_value=batch_size // self.minibatch_size,
            )
            if progress_bar
            else utils.NullProgressBar()
        )
        with progress:
            loss = self.run_batch(
                data,
                objective,
                extra_feeds=extra_feeds,
                callback=lambda *_: progress.step(),
                combine=combine,
                training=training,
            )

        # sum across objectives
        loss = np.sum(list(loss.values()))

        return loss

    def run_batch(
        self,
        data,
        outputs,
        extra_feeds=None,
        extra_fetches=None,
        n_epochs=1,
        truncation=None,
        shuffle=False,
        profile=False,
        training=False,
        callback=None,
        combine=np.stack,
        isolate_state=True,
    ):
        """
        Run the simulation on a batch of input data, computing the given
        output functions.

        Parameters
        ----------
        data : dict of {`~nengo.Node` or `~nengo.Probe`: \
                        `~numpy.ndarray`} or int
            Input values for Nodes in the network or target values for Probes;
            arrays should have shape ``(batch_size, n_steps,
            node.size_out/probe.size_in)``.  If no input data is required,
            an integer can be given specifying the number of timesteps to
            run the simulation.
        outputs : dict of {(tuple of) `~nengo.Probe`: callable or None}
            Functions to apply to probe outputs.  Functions can accept one
            positional argument (the output from that probe on one minibatch)
            or two (also passed the corresponding target value from ``data``).
            If a tuple of Probes are given as the key then the first
            argument will be a list of probe outputs, and the second
            argument will be the corresponding list of target values.  The
            function can return a ``tf.Tensor``, or tuple of Tensors,
            which will be evaluated on each minibatch of data.  If ``None``
            is given then the return value will be the output value from that
            probe.
        extra_feeds : dict of {``tf.Tensor``: `~numpy.ndarray`}
            Can be used to feed a value for arbitrary Tensors in the simulation
            (will be passed directly to the TensorFlow session)
        extra_fetches : (list/tuple/dict of) ``tf.Tensor``
            Can be used to fetch arbitrary (structures of) Tensor values from
            the simulation (will be fetched directly from the TensorFlow
            session).
        n_epochs : int
            Repeat ``data`` for ``n_epochs`` iterations.
        truncation : int
            If not None, run the simulation ``truncation`` timesteps at a time.
            Outputs from each truncation block will be passed sequentially to
            ``combine``, in the same way as minibatch blocks.  Note
            that the simulation state is preserved between truncation blocks,
            so the sequence forms one continuous run within each minibatch.
        shuffle : bool
            If True, randomize the data into different minibatches each epoch.
        profile : bool
            If True, collect TensorFlow profiling information while the
            simulation is running (this will slow down the simulation).
            Can also pass a string specifying a non-default filename for the
            saved profile data.
        training : bool
            If True, run the network in training mode, otherwise run it in
            inference mode (this can affect things like the neuron model
            used).
        callback : callable
            A function that will be called after each minibatch is evaluated.
            The function is passed two arguments; the first is a dictionary
            corresponding to ``outputs`` with the output values from each
            function, and the second is the value of ``extra_feeds``.
        combine : callable
            The function that will be used to combine the outputs from each
            minibatch/truncation block.  The values from each output function
            on each minibatch will be formed into a list and passed to
            ``combine`` in order to compute the final return values from
            this function.  Note that if the output function returns multiple
            values, then ``combine`` will be applied separately to each of
            those outputs across the minibatches.
        isolate_state : bool
            If True (default), isolate the simulation state for this run
            from the rest of the simulation (so the execution of this run
            is not affected by previous runs and will not affect future runs).
            If False, then this run begins from the terminal state of the
            last run, each minibatch will continue in sequence from the state
            of the previous, and future runs will resume from the terminal
            state of the last minibatch of this run.

        Returns
        -------
        output_vals : dict of {(tuple of) `~nengo.Probe`: \
                               (tuple of) `~numpy.ndarray`}
            The result of computing ``outputs`` on simulation probe values,
            given ``data``.  This pseudocode may help to understand how the
            return values are constructed given the various parameters of this
            function:

            .. code-block:: python

                output_vals = {}
                for probe, func in outputs.items():
                    probe_vals = []
                    for i in range(n_epochs):
                        for minibatch in data:
                            network_output = run_network(minibatch)
                            probe_vals.append(func(network_output[probe]))
                    output_vals[probe] = combine(output_values)

            Note that this is not how the values are computed in practice,
            as it would be quite inefficient.  This pseudocode also omits
            some of the finer details (e.g. truncation and state isolation).

        Notes
        -----
        In general, users should call one of the wrappers for this function
        (e.g., `.run_steps`, `.train`, or `.loss`),
        according to their use case.  However, this function can be called
        directly to run the simulation in a customized way.
        """

        n_steps = data if isinstance(data, int) else next(iter(data.values())).shape[1]

        # error checking
        if self.closed:
            raise SimulatorClosed("Simulator cannot run because it is closed.")
        if not isinstance(data, int):
            self._check_data(data)
        if n_steps % self.unroll != 0:
            raise ValidationError(
                "The number of timesteps in batch data must be evenly "
                "divisible by unroll_simulation",
                "data",
            )
        if truncation is not None and truncation % self.unroll != 0:
            raise ValidationError(
                "Truncation length must be evenly divisible by unroll_simulation",
                "truncation",
            )
        if training and self.tensor_graph.inference_only:
            raise ValidationError(
                "Network was created with inference_only=True, cannot "
                "be run in training mode",
                "inference_only",
            )

        if extra_fetches is None:
            extra_fetches = []

        # apply functions (if any) to output probes
        output_ops, init_ops = self.tensor_graph.build_outputs(outputs)

        # initialize any new variables
        if init_ops is not None:
            self.sess.run(init_ops)

        # save the internal state of the simulator
        if isolate_state:
            tmpdir = tempfile.TemporaryDirectory()
            self.save_params(
                os.path.join(tmpdir.name, "tmp"),
                include_local=True,
                include_global=False,
            )

        # set up profiling
        if profile:
            run_options = tf_compat.RunOptions(
                trace_level=tf_compat.RunOptions.FULL_TRACE
            )
            run_metadata = tf_compat.RunMetadata()
        else:
            run_options = None
            run_metadata = None

        # compute outputs on batch
        output_vals = collections.defaultdict(list)
        for _ in range(n_epochs):
            for offset, mini_data in utils.minibatch_generator(
                data,
                self.minibatch_size,
                truncation=truncation,
                shuffle=shuffle,
                rng=self.rng,
            ):
                if offset == 0 and isolate_state:
                    self.soft_reset()

                # fill in feed_dict values
                if isinstance(mini_data, int):
                    steps = mini_data
                    mini_data = None
                else:
                    steps = next(iter(mini_data.values())).shape[1]
                feed = self._fill_feed(
                    steps,
                    data=mini_data,
                    training=training,
                    start=offset + (0 if isolate_state else self.n_steps),
                )
                if extra_feeds is not None:
                    feed.update(extra_feeds)

                # run the simulation
                try:
                    out_vals, extra_vals = self.sess.run(
                        (output_ops, extra_fetches),
                        feed_dict=feed,
                        options=run_options,
                        run_metadata=run_metadata,
                    )
                except (tf.errors.InternalError, tf.errors.UnknownError) as e:
                    if e.op is not None and e.op.type == "PyFunc":
                        raise SimulationError(
                            "Function '%s' caused an error (see error log "
                            "above)" % e.op.name
                        )
                    else:
                        raise e  # pragma: no cover (unknown errors)

                if callback is not None:
                    callback(out_vals, extra_vals)

                for k, v in out_vals.items():
                    output_vals[k].append(v)

        # restore internal state of simulator
        if isolate_state:
            self.load_params(
                os.path.join(tmpdir.name, "tmp"),
                include_local=True,
                include_global=False,
            )
            tmpdir.cleanup()

        # combine outputs from each minibatch
        for probe, vals in output_vals.items():
            # if the output function returns multiple items, keep those
            # arrays separate
            if isinstance(vals[0], (list, tuple)):
                output_vals[probe] = tuple(combine(v) for v in zip(*vals))
            else:
                output_vals[probe] = combine(vals)

        # convert back from defaultdict
        output_vals = dict(output_vals)

        # output profile data to file
        self._profile_output(profile, run_metadata)

        return output_vals

    def save_params(self, path, include_global=True, include_local=False):
        """
        Save network parameters to the given ``path``.

        Parameters
        ----------
        path : str
            Filepath of parameter output file
        include_global : bool
            If True (default True), save global/trainable network variables
        include_local : bool
            If True (default False), save local (non-trainable) network
            variables

        Notes
        -----
        This function is useful for saving/loading entire models; for
        saving/loading individual objects within a model, see
        `.get_nengo_params`.
        """
        if self.closed:
            raise SimulatorClosed("Simulation has been closed, cannot save parameters")

        with self.tensor_graph.graph.as_default():
            vars = []
            if include_global:
                vars.extend(tf_compat.global_variables())
            if include_local:
                vars.extend(tf_compat.local_variables())

            with tf.device("/cpu:0"):
                path = tf_compat.train.Saver(vars).save(self.sess, path)

        logger.info("Model parameters saved to %s", path)

    def load_params(self, path, include_global=True, include_local=False):
        """
        Load network parameters from the given ``path``.

        Parameters
        ----------
        path : str
            Filepath of parameter input file
        include_global : bool
            If True (default True), load global (trainable) network variables
        include_local : bool
            If True (default False), load local (non-trainable) network
            variables

        Notes
        -----
        This function is useful for saving/loading entire models; for
        saving/loading individual objects within a model, see
        `.get_nengo_params`.
        """
        if self.closed:
            raise SimulatorClosed("Simulation has been closed, cannot load parameters")

        with self.tensor_graph.graph.as_default():
            vars = []
            if include_global:
                vars.extend(tf_compat.global_variables())
            if include_local:
                vars.extend(tf_compat.local_variables())

            with tf.device("/cpu:0"):
                tf_compat.train.Saver(vars).restore(self.sess, path)

        logger.info("Model parameters loaded from %s", path)

    def freeze_params(self, objs):
        """
        Stores the live parameter values from the simulation back into a
        Nengo object definition.

        This can be helpful for reusing a NengoDL model inside a different
        Simulator.  For example:

        .. code-block:: python

            with nengo.Network() as net:
                < build network >

            with nengo_dl.Simulator(net) as sim:
                < run some optimization >
                sim.freeze_params(net)

            with nengo.Simulator(net) as sim2:
                # run the network in the default Nengo simulator, with the
                # trained parameters
                sim2.run(1.0)

        Parameters
        ----------
        obj : (list of) ``NengoObject``
            The Nengo object(s) into which parameter values will be stored.
            Note that these objects must be members of the Network used to
            initialize the Simulator.

        Notes
        -----
        This modifies the source object in-place, and it may slightly modify
        the structure of that object.  The goal is to have the object produce
        the same output as it would if run in the NengoDL simulator.  It may
        not be possible to accurately freeze all possible object; if you run
        into errors in this process, try manually extracting the parameters you
        need in your model (from ``sim.data``).
        """

        if self.closed:
            raise SimulatorClosed(
                "Simulation has been closed, cannot freeze parameters"
            )

        if not isinstance(objs, (list, tuple)):
            objs = [objs]

        for obj in objs:
            if obj not in [self.model.toplevel] + self.model.toplevel.all_objects:
                raise ValueError(
                    "%s is not a member of the Network used to "
                    "initialize the Simulator"
                )

            if not isinstance(obj, (Network, Ensemble, Connection)):
                raise TypeError(
                    "Objects of type %s do not have parameters to store" % type(obj)
                )

            if isinstance(obj, Network):
                todo = obj.all_ensembles + obj.all_connections
            else:
                todo = [obj]

            for o, params in zip(todo, self.get_nengo_params(todo)):
                for k, v in params.items():
                    setattr(o, k, v)

    def get_nengo_params(self, nengo_objs, as_dict=False):
        """
        Extract model parameters in a form that can be used to initialize
        Nengo objects in a different model.

        For example:

        .. code-block:: python

            with nengo.Network() as net:
                a = nengo.Ensemble(10, 1)
                b = nengo.Ensemble(10, 1)
                c = nengo.Connection(a, b)

            with nengo_dl.Simulator(net) as sim:
                # < do some optimization >
                params = sim.get_nengo_params([a, b, c])

            with nengo.Network() as new_net:
                # < build some other network >

                # now we want to insert two connected ensembles with
                # the same parameters as our previous network:
                d = nengo.Ensemble(10, 1, **params[0])
                e = nengo.Ensemble(10, 1, **params[1])
                f = nengo.Connection(d, e, **params[2])

        Parameters
        ----------
        nengo_objs : (list of) `~nengo.Ensemble` or `~nengo.Connection`
            A single object or list of objects for which we want to get the
            parameters.
        as_dict : bool
            If True, return the values as a dictionary keyed by object label,
            instead of a list (the default).  Note that in this case labels
            must be unique.

        Returns
        -------
        params : (list or dict) of dicts
            kwarg dicts corresponding to ``nengo_objs`` (passing these
            dicts as kwargs when creating new Nengo objects will result in a
            new object with the same parameters as the source object).  A
            single kwarg dict if a single object was passed in, or a list
            (dict if ``as_dict=True``) of kwargs corresponding to multiple
            input objects.
        """

        if isinstance(nengo_objs, (list, tuple)):
            scalar = False
        else:
            scalar = True
            nengo_objs = [nengo_objs]

        # convert neurons to the parent ensemble
        nengo_objs = [
            obj.ensemble if isinstance(obj, Neurons) else obj for obj in nengo_objs
        ]

        # find all the data we need to fetch
        fetches = []
        for obj in nengo_objs:
            if isinstance(obj, Connection):
                fetches.append((obj, "weights"))
            elif isinstance(obj, Ensemble):
                if isinstance(obj.neuron_type, Direct):
                    # we cannot transfer direct ensemble parameters, because
                    # the nengo builder ignores the encoders specified for
                    # a direct ensemble
                    raise ValueError(
                        "get_nengo_params will not work correctly for "
                        "Direct neuron ensembles. Try manually translating "
                        "your network using `sim.data` instead."
                    )

                fetches.extend([(obj, "scaled_encoders"), (obj, "bias")])
            else:
                raise ValueError(
                    "Can only get Nengo parameters for Ensembles or Connections"
                )

        # get parameter values from simulation
        data = self.data.get_params(*fetches)

        # store parameter values in a form that can be loaded in nengo
        params = []
        idx = 0
        for obj in nengo_objs:
            if isinstance(obj, Connection):
                weights = data[idx]
                idx += 1
                if isinstance(obj.pre_obj, Ensemble):
                    params.append(
                        {
                            "solver": NoSolver(weights.T, weights=False),
                            "function": lambda x, weights=weights: np.zeros(
                                weights.shape[0]
                            ),
                            "transform": 1,
                        }
                    )
                elif isinstance(obj.transform, Convolution):
                    transform = copy.copy(obj.transform)
                    # manually bypass the read-only check (we are sure that
                    # nothing else has a handle to the new transform at this
                    # point, so this won't cause any problems)
                    Convolution.init.data[transform] = weights
                    params.append({"transform": transform})
                else:
                    if all(x == 1 for x in weights.shape):
                        weights = np.squeeze(weights)
                    params.append({"transform": weights})
            else:
                # note: we don't want to change the original gain (even though
                # it is rolled into the encoder values), because connections
                # direct to `ens.neurons` will still use the gains (and those
                # gains are not updated during training, only the encoders)
                gain = self.model.params[obj].gain

                # the encoders we get from the simulation are the actual
                # weights we want in the simulation. but during the build
                # process, gains and radius will be applied to the encoders.
                # so we need to undo that scaling here, so that the build
                # process will result in the correct values.
                encoders = data[idx] * obj.radius / gain[:, None]

                params.append(
                    {
                        "encoders": encoders,
                        "normalize_encoders": False,
                        "gain": gain,
                        "bias": data[idx + 1],
                        "max_rates": Ensemble.max_rates.default,
                        "intercepts": Ensemble.intercepts.default,
                    }
                )
                idx += 2

        # return params in appropriate format
        if scalar:
            return params[0]

        if as_dict:
            param_dict = {}
            for obj, p in zip(nengo_objs, params):
                if obj.label in param_dict:
                    raise ValueError(
                        "Duplicate label ('%s') detected; cannot return "
                        "parameters with as_dict=True" % obj.label
                    )
                else:
                    param_dict[obj.label] = p
            params = param_dict

        return params

    def check_gradients(self, outputs=None, atol=1e-5, rtol=1e-3):
        """
        Perform gradient checks for the network (used to verify that the
        analytic gradients are correct).

        Raises a simulation error if the difference between analytic and
        numeric gradient is greater than ``atol + rtol * numeric_grad``
        (elementwise).

        Parameters
        ----------
        outputs : ``tf.Tensor`` or list of ``tf.Tensor`` or \
                  list of `~nengo.Probe`
            Compute gradients wrt this output (if None, computes wrt each
            output probe)
        atol : float
            Absolute error tolerance
        rtol : float
            Relative (to numeric grad) error tolerance

        Notes
        -----
        Calling this function will reset all values in the network, so it
        should not be intermixed with calls to `.Simulator.run`.
        """

        if self.tensor_graph.inference_only:
            raise ValidationError(
                "Network was created with inference_only=True, cannot "
                "compute gradients",
                "inference_only",
            )

        delta = 1e-3
        n_steps = self.unroll * 2

        data = {
            n: np.zeros((self.minibatch_size, n_steps, n.size_out))
            for n in self.tensor_graph.invariant_inputs
        }
        data.update(
            {
                p: np.zeros((self.minibatch_size, n_steps, p.size_in))
                for p in self.tensor_graph.target_phs
            }
        )
        feed = self._fill_feed(n_steps, data=data, training=True)

        if outputs is None:
            # note: the x + 0 is necessary because `gradient_checker`
            # doesn't work properly if the output variable is a tensorarray
            outputs = [x + 0 for x in self.tensor_graph.probe_arrays.values()]
        elif isinstance(outputs, tf.Tensor):
            outputs = [outputs]
        else:
            outputs = [self.tensor_graph.probe_arrays[p] + 0 for p in outputs]

        # check gradient wrt inp
        for node, inp in self.tensor_graph.input_ph.items():
            inp_shape = inp.get_shape().as_list()
            inp_shape = [n_steps if x is None else x for x in inp_shape]
            inp_tens = self.tensor_graph.input_ph[node]
            feed[inp_tens] = np.ascontiguousarray(feed[inp_tens])
            inp_val = np.ravel(feed[inp_tens])
            for out in outputs:
                out_shape = out.get_shape().as_list()
                out_shape = [n_steps if x is None else x for x in out_shape]

                # we need to compute the numeric jacobian manually, to
                # correctly handle variables (tensorflow doesn't expect
                # state ops in `compute_gradient`, because it doesn't define
                # gradients for them)
                numeric = np.zeros(
                    (
                        np.prod(inp_shape, dtype=np.int32),
                        np.prod(out_shape, dtype=np.int32),
                    )
                )

                for i in range(numeric.shape[0]):
                    self.soft_reset()
                    inp_val[i] = delta
                    plus = self.sess.run(out, feed_dict=feed)

                    self.soft_reset()
                    inp_val[i] = -delta
                    minus = self.sess.run(out, feed_dict=feed)

                    numeric[i] = np.ravel((plus - minus) / (2 * delta))

                    inp_val[i] = 0

                self.soft_reset()

                with tf_compat.variable_scope(tf_compat.get_variable_scope()) as scope:
                    dx, dy = gradient_checker._compute_dx_and_dy(inp, out, out_shape)

                    self.sess.run(
                        tf_compat.variables_initializer(
                            scope.get_collection("gradient_vars")
                        )
                    )

                with self.sess.as_default():
                    analytic = gradient_checker._compute_theoretical_jacobian(
                        inp,
                        inp_shape,
                        np.zeros(inp_shape),
                        dy,
                        out_shape,
                        dx,
                        extra_feed_dict=feed,
                    )

                if np.any(np.isnan(analytic)) or np.any(np.isnan(numeric)):
                    raise SimulationError("NaNs detected in gradient")
                fail = abs(analytic - numeric) >= atol + rtol * abs(numeric)
                if np.any(fail):
                    raise SimulationError(
                        "Gradient check failed for input %s and output %s\n"
                        "numeric values:\n%s\n"
                        "analytic values:\n%s\n"
                        % (node, out, numeric[fail], analytic[fail])
                    )

        self.soft_reset()

        logger.info("Gradient check passed")

    def trange(self, sample_every=None, dt=None):
        """
        Create a vector of times matching probed data.

        Note that the range does not start at 0 as one might expect, but at
        the first timestep (i.e., ``dt``).

        Parameters
        ----------
        sample_every : float (Default: None)
            The sampling period of the probe to create a range for.
            If None, a time value for every ``dt`` will be produced.
        """

        if dt is not None:
            if sample_every is not None:
                raise ValidationError(
                    "Cannot specify both `dt` and `sample_every`. "
                    "Use `sample_every` only.",
                    attr="dt",
                    obj=self,
                )
            warnings.warn("`dt` is deprecated. Use `sample_every` instead.")
            sample_every = dt

        period = 1 if sample_every is None else sample_every / self.dt
        steps = np.arange(1, self.n_steps + 1)
        return self.dt * steps[steps % period < 1]

    def close(self):
        """
        Close the simulation, freeing resources.

        Notes
        -----
        The simulation cannot be restarted after it is closed.  This is not a
        technical limitation, just a design decision made for all Nengo
        simulators.
        """

        if not self.closed:
            # note: we use getattr in case it crashes before the object is
            # created
            if getattr(self, "sess", None) is not None:
                self.sess.close()
            self.sess = None

            if getattr(self, "summary", None) is not None:
                self.summary.close()

            self.closed = True

    def _fill_feed(self, n_steps, data=None, start=0, training=False):
        """
        Create a feed dictionary containing values for all the placeholder
        inputs in the network, which will be passed to ``tf.Session.run``.

        Parameters
        ----------
        n_steps : int
            The number of execution steps
        data : dict of {`~nengo.Node` or `~nengo.Probe` : `~numpy.ndarray`}
            Input values for Nodes and target values for Probes.  Arrays
            should have shape ``(sim.minibatch_size, n_steps,
            node.size_out/probe.size_in)``.
        start : int
            Initial value of simulator timestep
        training : bool
            Whether we are running in training or inference mode

        Returns
        -------
        feed_dict : dict of {``tf.Tensor``: `~numpy.ndarray`}
            Feed values for placeholder tensors in the network
        """

        # fill in constants
        feed_dict = {
            self.tensor_graph.step_var: start,
            self.tensor_graph.stop_var: start + n_steps,
        }
        if not self.tensor_graph.inference_only:
            feed_dict[self.tensor_graph.signals.training] = training

        inputs = {}
        targets = {}
        if data is not None:
            for k, v in data.items():
                if isinstance(k, Node):
                    inputs[k] = v
                elif isinstance(k, Probe):
                    targets[k] = v
                else:
                    # this should be caught in check_data
                    raise NotImplementedError()

        # fill in input values
        feed_dict.update(self._generate_inputs(inputs, n_steps))

        # fill in target values
        for p, t in targets.items():
            if p not in self.tensor_graph.target_phs:
                raise ValidationError(
                    "%s is not a valid target; this is probably because "
                    "it is not used in the objective function" % p,
                    "targets",
                )

            feed_dict[self.tensor_graph.target_phs[p]] = t

        return feed_dict

    def _generate_inputs(self, data, n_steps):
        """
        Generate inputs for the network (the output values of each Node with
        no incoming connections).

        Parameters
        ----------
        data : dict of {`~nengo.Node`: `~numpy.ndarray`}
            Override the values of input Nodes with the given data.  Arrays
            should have shape ``(sim.minibatch_size, n_steps, node.size_out)``.
        n_steps : int
            Number of simulation timesteps for which to generate input data

        Returns
        -------
        feed_vals : dict of {`~nengo.Node`: `~numpy.ndarray}
            Simulation values for all the input Nodes in the network.
        """

        feed_vals = {}
        for n, output in self.tensor_graph.input_funcs.items():
            if n in data:
                # move minibatch dimension to the end
                feed_val = np.moveaxis(data[n], 0, -1)
            elif isinstance(output, np.ndarray):
                # tile to n_steps/minibatch size
                feed_val = np.tile(
                    output[None, :, None], (n_steps, 1, self.minibatch_size)
                )
            else:
                # call output function to determine value
                feed_val = np.zeros(
                    (n_steps, n.size_out, self.minibatch_size),
                    dtype=self.tensor_graph.dtype.as_numpy_dtype,
                )

                for i in range(n_steps):
                    # note: need to copy the output of func, as func
                    # may mutate its outputs in-place on subsequent calls.
                    # this assignment will broadcast the output along the
                    # minibatch dimension if required.
                    feed_val[i] = np.transpose(
                        [func((i + self.n_steps + 1) * self.dt) for func in output]
                    )

            # note: we still call the function (above) even if the output
            # is not being used, because it may have side-effects
            if n in self.tensor_graph.input_ph:
                feed_vals[self.tensor_graph.input_ph[n]] = feed_val

        return feed_vals

    def _check_data(self, data, n_batch=None, n_steps=None):
        """
        Performs error checking on simulation data.

        Parameters
        ----------
        data : dict of {`~nengo.Node` or `~nengo.Probe`: `~numpy.ndarray`}
            Array of data associated with given objects in model (Nodes or
            Probes)
        n_batch : int
            Number of elements in batch (if None, will just verify that all
            data items have same batch size)
        n_steps : int
            Number of simulation steps (if None, will just verify that all
            data items have same number of steps)
        """

        for d, x in data.items():
            if x.ndim != 3:
                raise ValidationError(
                    "should have rank 3 (batch_size, n_steps, dimensions), "
                    "found rank %d" % x.ndim,
                    "data",
                )

            if isinstance(d, Node):
                if d not in self.tensor_graph.invariant_inputs:
                    raise ValidationError(
                        "%s is not an input Node (a nengo.Node with "
                        "size_in==0), or is from a different network." % d,
                        "data",
                    )
            elif isinstance(d, Probe):
                if d not in self.model.probes:
                    raise ValidationError("%s is from a different network" % d, "data")
            else:
                raise ValidationError(
                    "Data objects must be Nodes or Probes, not %s" % d, "data"
                )

        args = [n_batch, n_steps]
        labels = ["batch size", "number of timesteps"]

        for i in range(2):
            if args[i] is None:
                val = next(iter(data.values())).shape[i]
                for n, x in data.items():
                    if x.shape[i] != val:
                        raise ValidationError(
                            "Elements have different %s: %s vs %s"
                            % (labels[i], val, x.shape[0]),
                            "data",
                        )
            else:
                for n, x in data.items():
                    if x.shape[i] != args[i]:
                        raise ValidationError(
                            "Data for %s has %s=%s, which does not match "
                            "expected size (%s)" % (n, labels[i], x.shape[i], args[i]),
                            "data",
                        )

        for n, x in data.items():
            if x.shape[0] < self.minibatch_size:
                raise ValidationError(
                    "Size of minibatch (%d) for %s data less than Simulation "
                    "`minibatch_size` (%d)" % (x.shape[0], n, self.minibatch_size),
                    "data",
                )

            d = n.size_out if isinstance(n, Node) else n.size_in
            if x.shape[2] != d:
                raise ValidationError(
                    "Dimensionality of data (%s) does not match "
                    "dimensionality of %s (%s)" % (x.shape[2], n, d),
                    "data",
                )

    def _profile_output(self, profile, run_metadata):
        """
        Outputs profile information to file.

        Parameters
        ----------
        profile : bool or str
            If True or a string (filename), output profile information to file
        run_metadata : ``tf.RunMetadata``
            TensorFlow RunMetadata proto populated with profiling data
        """

        if not profile:
            return

        trace = Timeline(step_stats=run_metadata.step_stats)
        if isinstance(profile, str):
            filename = profile
        else:
            filename = "nengo_dl_profile.json"
        with open(filename, "w") as f:
            f.write(trace.generate_chrome_trace_format())

    @property
    def dt(self):
        """The time (in seconds) represented by one simulation timestep."""
        return self.model.dt

    @dt.setter
    def dt(self, _):
        raise ReadonlyError(attr="dt", obj=self)

    @property
    def training_step(self):
        """The number of training iterations that have been executed."""
        return self.tensor_graph.training_step

    def __enter__(self):
        self._graph_context = self.tensor_graph.graph.as_default()
        self._device_context = self.tensor_graph.graph.device(self.tensor_graph.device)
        self._graph_context.__enter__()
        self._device_context.__enter__()
        self.sess.__enter__()
        return self

    def __exit__(self, *args):
        self.sess.__exit__(*args)
        self._device_context.__exit__(*args)
        self._graph_context.__exit__(*args)
        self.close()

    def __del__(self):
        """
        Raise a RuntimeWarning if the Simulator is deallocated while open.
        """

        if self.closed is not None and not self.closed:
            warnings.warn(
                "Simulator with model=%s was deallocated while open. "
                "Simulators should be closed manually to ensure resources "
                "are properly freed." % self.model,
                RuntimeWarning,
            )
            self.close()

    def __getstate__(self):
        raise NotImplementedError(
            "TensorFlow does not support pickling; see "
            "https://www.nengo.ai/nengo-dl/training.html"
            "#saving-and-loading-parameters "
            "for information on how to save/load a NengoDL model."
        )


class SimulationData(collections.Mapping):
    """
    Data structure used to access simulation data from the model.

    The main use case for this is to access Probe data; for example,
    ``probe_data = sim.data[my_probe]``.  However, it is also
    used to access the parameters of objects in the model; for example, after
    the model has been optimized via `.Simulator.train`, the updated
    encoder values for an ensemble can be accessed via
    ``trained_encoders = sim.data[my_ens].encoders``.

    Parameters
    ----------
    sim : `.Simulator`
        The simulator from which data will be drawn
    minibatched : bool
        If False, discard the minibatch dimension on probe data

    Notes
    -----
    SimulationData shouldn't be created/accessed directly by the user, but
    rather via ``sim.data`` (which is an instance of SimulationData).
    """

    def __init__(self, sim, minibatched):
        self.sim = sim
        self.minibatched = minibatched

    def __getitem__(self, obj):
        """Return the data associated with ``obj``.

        Parameters
        ----------
        obj : `~nengo.Probe` or `~nengo.Ensemble` or `~nengo.Connection`
            Object whose simulation data is being accessed

        Returns
        -------
        data : `~numpy.ndarray` or \
               `~nengo.builder.ensemble.BuiltEnsemble` or \
               `~nengo.builder.connection.BuiltConnection`
            Array containing probed data if ``obj`` is a
            `~nengo.Probe`, otherwise the corresponding
            parameter object
        """

        if obj not in self.sim.model.params:
            raise ValidationError(
                "Object is not in parameters of model %s" % self.sim.model, str(obj)
            )

        data = self.sim.model.params[obj]

        if isinstance(obj, Probe):
            if len(data) == 0:
                return []
            data = np.concatenate(data, axis=1)
            if not self.minibatched:
                data = data[0]

            data.setflags(write=False)
        elif isinstance(obj, Ensemble):
            if isinstance(obj.neuron_type, Direct):
                # direct mode ensemble
                gain = bias = None
                scaled_encoders = encoders = self.get_params((obj, "scaled_encoders"))[
                    0
                ]
            else:
                # get the live simulation values
                scaled_encoders, bias = self.get_params(
                    (obj, "scaled_encoders"), (obj, "bias")
                )

                # infer the related values (rolled into scaled_encoders)
                gain = (
                    obj.radius
                    * np.linalg.norm(scaled_encoders, axis=-1)
                    / np.linalg.norm(data.encoders, axis=-1)
                )
                encoders = obj.radius * scaled_encoders / gain[:, None]

            # figure out max_rates/intercepts from neuron model
            max_rates, intercepts = obj.neuron_type.max_rates_intercepts(gain, bias)

            data = BuiltEnsemble(
                data.eval_points,
                encoders,
                intercepts,
                max_rates,
                scaled_encoders,
                gain,
                bias,
            )
        elif isinstance(obj, Connection):
            # get the live simulation values
            weights = self.get_params((obj, "weights"))[0]

            # impossible to recover transform
            transform = None

            data = BuiltConnection(
                data.eval_points, data.solver_info, weights, transform
            )

        return data

    def get_params(self, *obj_attrs):
        """
        Returns the current parameter values for the given objects.

        Parameters
        ----------
        obj_attrs : list of (``NengoObject``, str)
            The Nengo object and attribute of that object for which we want
            to know the parameter values (each object-attribute pair specified
            as a tuple argument to the function).

        Returns
        -------
        params : list of `~numpy.ndarray`
            Current values of the requested parameters

        Notes
        -----
        Parameter values should be accessed through ``sim.data``
        (which will call this function if necessary), rather than directly
        through this function.
        """

        if self.sim.closed:
            warnings.warn(
                "Checking parameters after simulator is closed; "
                "cannot fetch live values, so the initial values "
                "will be returned."
            )

            return [
                getattr(self.sim.model.params[obj], attr) for obj, attr in obj_attrs
            ]

        params = []
        sigs = []
        fetches = {}
        for obj, attr in obj_attrs:

            sig_obj, sig_attr = self._attr_map(obj, attr)
            sig = self.sim.model.sig[sig_obj][sig_attr]
            sigs.append(sig)

            if sig not in self.sim.tensor_graph.signals:
                # if sig isn't in sig_map then that means it isn't used
                # anywhere in the simulation (and therefore never changes), so
                # we can safely return the static build value
                params.append(getattr(self.sim.model.params[obj], attr))
            else:
                # this is a live parameter value we need to fetch from the
                # simulation. we queue them up and fetch them all at once to
                # be more efficient
                placeholder = object()
                fetches[placeholder] = self.sim.tensor_graph.get_tensor(sig)
                params.append(placeholder)

        # get the live parameter values
        fetched = self.sim.sess.run(fetches)

        # final updating of parameters
        for i, sig in enumerate(sigs):
            # fill in placeholder values
            if type(params[i]) == object:
                params[i] = fetched[params[i]]

            # handle minibatch dimension
            if sig.minibatched:
                if not self.minibatched:
                    params[i] = params[i][..., 0]
                else:
                    params[i] = np.moveaxis(params[i], -1, 0)

        return params

    def _attr_map(self, obj, attr):
        """
        Maps from ``sim.data[obj].attr`` to the equivalent
        ``model.sig[obj][attr]``.

        Parameters
        ----------
        obj : ``NengoObject``
            The nengo object for which we want to know the parameters
        attr : str
            The parameter of ``obj`` to be returned

        Returns
        -------
        obj : ``NengoObject``
            The nengo object to key into ``model.sig``
        attr : str
            The name of the signal corresponding to input attr

        """

        if isinstance(obj, Ensemble) and attr == "bias":
            return obj.neurons, attr
        elif isinstance(obj, Ensemble) and attr == "scaled_encoders":
            return obj, "encoders"

        return obj, attr

    def __len__(self):
        return len(self.sim.model.params)

    def __iter__(self):
        return iter(self.sim.model.params)



# pylint: disable=missing-docstring

import nengo
import numpy as np
import pytest
import tensorflow as tf

from nengo_dl import objectives
from nengo_dl.compat import tf_compat


@pytest.mark.parametrize(
    "axis, weight, order", [(None, 0.6, 1), (None, 0.7, "euclidean"), (1, None, 2)]
)
def test_regularize(axis, weight, order, rng, sess):
    x_init = rng.randn(2, 3, 4, 5)

    x = tf.constant(x_init)

    reg = objectives.Regularize(weight=weight, order=order, axis=axis)(x)

    sess.run(tf_compat.global_variables_initializer())
    reg_val = sess.run(reg)

    if order == "euclidean":
        order = 2

    if weight is None:
        weight = 1

    if axis is None:
        truth = np.reshape(x_init, x_init.shape[:2] + (-1,))
        axis = 2
    else:
        truth = x_init
        axis += 2
    truth = np.mean(np.linalg.norm(truth, ord=order, axis=axis))
    truth *= weight

    assert np.allclose(reg_val, truth)


@pytest.mark.training
@pytest.mark.parametrize("mode", ("activity", "weights"))
def test_regularize_train(Simulator, mode, seed):
    with nengo.Network(seed=seed) as net:
        a = nengo.Node([1])
        b = nengo.Ensemble(
            30,
            1,
            neuron_type=nengo.Sigmoid(tau_ref=1),
            gain=nengo.dists.Choice([1]),
            bias=nengo.dists.Choice([0]),
        )
        c = nengo.Connection(
            a, b.neurons, synapse=None, transform=nengo.dists.Uniform(-0.1, 0.1)
        )

        if mode == "weights":
            p = nengo.Probe(c, "weights")
        else:
            p = nengo.Probe(b.neurons)

    with Simulator(net) as sim:
        sim.train(
            5,
            tf_compat.train.RMSPropOptimizer(0.01 if mode == "weights" else 0.1),
            objective={p: objectives.Regularize()},
            n_epochs=100,
        )

        sim.step()
        assert np.allclose(sim.data[p], 0, atol=1e-2)



"""
Some common objective functions (for use with the ``objective`` argument in
`.Simulator.train` or `.Simulator.loss`).
"""

import tensorflow as tf

from nengo_dl.compat import tf_math


def mse(outputs, targets):
    """
    Compute Mean Squared Error between given outputs and targets.

    If any values in ``targets`` are ``nan``, that will be treated as
    zero error for those elements.

    Parameters
    ----------
    outputs : ``tf.Tensor``
        Output values from a Probe in a network.
    targets : ``tf.Tensor``
        Target values for a Probe in a network.

    Returns
    -------
    mse : ``tf.Tensor``
        Tensor representing the mean squared error.
    """

    targets = tf.where(tf_math.is_nan(targets), outputs, targets)
    return tf.reduce_mean(input_tensor=tf.square(targets - outputs))


class Regularize:
    """
    An objective function to apply regularization penalties.

    Parameters
    ----------
    order : int or str
        Order of the regularization norm (e.g. ``1`` for L1 norm, ``2`` for
        L2 norm).  See https://www.tensorflow.org/api_docs/python/tf/norm for
        a full description of the possible values for this parameter.
    axis : int or None
        The axis of the probed signal along which to compute norm.  If None
        (the default), the signal is flattened and the norm is computed across
        the resulting vector.  Note that these are only the axes with respect
        to the output on a single timestep (i.e. batch/time dimensions are not
        included).
    weight : float
        Scaling weight to apply to regularization penalty.

    Notes
    -----
    The mean will be computed across all the non-``axis`` dimensions after
    computing the norm (including batch/time) in order to compute the overall
    objective value.
    """

    def __init__(self, order=2, axis=None, weight=None):
        self.order = order
        self.axis = axis
        self.weight = weight

    def __call__(self, x):
        if self.axis is None:
            if x.get_shape().ndims > 3:
                # flatten signal (keeping batch/time dimension)
                x = tf.reshape(x, tf.concat([tf.shape(input=x)[:2], (-1,)], axis=0))
            axis = 2
        else:
            axis = self.axis + 2

        output = tf.reduce_mean(
            input_tensor=tf.norm(tensor=x, axis=axis, ord=self.order)
        )

        if self.weight is not None:
            output *= self.weight

        return output



# pylint: disable=missing-docstring

from nengo import ensemble, Network, Ensemble, Connection, Probe
from nengo.builder import Model
from nengo.exceptions import ValidationError, ConfigError, NetworkContextError
import numpy as np
import pytest

from nengo_dl import config, builder


def test_configure_trainable():
    with Network() as net:
        conf = net.config
        config.configure_settings(trainable=None)

    assert conf[Ensemble].trainable is None
    assert conf[Connection].trainable is None
    assert conf[ensemble.Neurons].trainable is None

    # check that we can set trainable after it is set up for configuration
    conf[Ensemble].trainable = True

    # check that boolean value is enforced
    with pytest.raises(ValidationError):
        conf[Ensemble].trainable = 5

    assert conf[Ensemble].trainable is True

    # check that calling configure again overrides previous changes
    with net:
        config.configure_settings(trainable=None)

    assert conf[Ensemble].trainable is None

    # check that non-None defaults work
    with net:
        config.configure_settings(trainable=False)

    assert conf[Ensemble].trainable is False

    # check that calling configure outside network context is an error
    with pytest.raises(NetworkContextError):
        config.configure_settings(trainable=None)

    # check that passing an invalid parameter raises an error
    with net:
        with pytest.raises(ConfigError):
            config.configure_settings(troinable=None)


@pytest.mark.parametrize("as_model", (True, False))
def test_session_config(Simulator, as_model):
    with Network() as net:
        config.configure_settings(
            session_config={
                "graph_options.optimizer_options.opt_level": 21,
                "gpu_options.allow_growth": True,
            }
        )

    if as_model:
        # checking that config settings work when we pass in a model instead of
        # network
        model = Model(dt=0.001, builder=builder.NengoBuilder())
        model.build(net)
        net = None
    else:
        model = None

    with Simulator(net, model=model) as sim:
        assert sim.sess._config.graph_options.optimizer_options.opt_level == 21
        assert sim.sess._config.gpu_options.allow_growth


def test_keep_history(Simulator, seed):
    with Network(seed=seed) as net:
        config.configure_settings(keep_history=True)
        a = Ensemble(30, 1)
        p = Probe(a.neurons, synapse=0.1)

    with Simulator(net) as sim:
        sim.run_steps(10)

    with net:
        net.config[p].keep_history = False

    with Simulator(net) as sim2:
        sim2.run_steps(10)

    assert sim.data[p].shape == (10, 30)
    assert sim2.data[p].shape == (1, 30)
    assert np.allclose(sim.data[p][[-1]], sim2.data[p])



# pylint: disable=missing-docstring

from collections import OrderedDict
from distutils.version import LooseVersion
import logging
import os
import pickle

import nengo
from nengo.dists import Uniform
from nengo.exceptions import (
    SimulationError,
    SimulatorClosed,
    ReadonlyError,
    ValidationError,
)
import numpy as np
import pytest
import tensorflow as tf

from nengo_dl import configure_settings, tensor_layer, dists, TensorNode
from nengo_dl.compat import tf_compat, tf_uniform
from nengo_dl.objectives import mse
from nengo_dl.simulator import SimulationData
from nengo_dl.tests import dummies
from nengo_dl.utils import tf_gpu_installed


def test_persistent_state(Simulator, seed):
    """Make sure that state is preserved between runs."""

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node(np.sin)
        ens = nengo.Ensemble(1000, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    with Simulator(net) as sim:
        sim.run_steps(100)
        data = sim.data[p]
        sim.reset()

        sim.run_steps(100)
        data2 = sim.data[p]
        sim.reset()

        for _ in range(100 // sim.unroll):
            sim.run_steps(sim.unroll)
        data3 = sim.data[p]

    assert np.allclose(data, data2)
    assert np.allclose(data2, data3)


def test_step_blocks(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        inp = nengo.Node(np.sin)
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    with Simulator(net, unroll_simulation=25) as sim1:
        sim1.run_steps(50)
    with Simulator(net, unroll_simulation=10) as sim2:
        sim2.run_steps(50)

    assert np.allclose(sim1.data[p], sim2.data[p])

    with pytest.warns(RuntimeWarning):
        with Simulator(net, unroll_simulation=5) as sim:
            sim.run_steps(2)


def test_minibatch(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        inp = [
            nengo.Node(output=[0.5]),
            nengo.Node(output=np.sin),
            nengo.Node(output=nengo.processes.WhiteSignal(5, 0.5, seed=seed)),
        ]

        ens = [
            nengo.Ensemble(10, 1, neuron_type=nengo.AdaptiveLIF()),
            nengo.Ensemble(10, 1, neuron_type=nengo.LIFRate()),
            nengo.Ensemble(10, 2, noise=nengo.processes.WhiteNoise(seed=seed)),
        ]

        nengo.Connection(inp[0], ens[0])
        nengo.Connection(inp[1], ens[1], synapse=None)
        nengo.Connection(inp[2], ens[2], synapse=nengo.Alpha(0.1), transform=[[1], [1]])
        conn = nengo.Connection(ens[0], ens[1], learning_rule_type=nengo.PES())
        nengo.Connection(inp[0], conn.learning_rule)

        ps = [nengo.Probe(e) for e in ens]

    with Simulator(net, minibatch_size=None) as sim:
        probe_data = [[] for _ in ps]
        for _ in range(5):
            sim.run_steps(100)

            for j, p in enumerate(ps):
                probe_data[j] += [sim.data[p]]

            sim.reset()

        probe_data = [np.stack(x, axis=0) for x in probe_data]

    with Simulator(net, minibatch_size=5) as sim:
        sim.run_steps(100)

    assert np.allclose(sim.data[ps[0]], probe_data[0], atol=1e-6)
    assert np.allclose(sim.data[ps[1]], probe_data[1], atol=1e-6)
    assert np.allclose(sim.data[ps[2]], probe_data[2], atol=1e-6)


def test_input_feeds(Simulator):
    minibatch_size = 10

    with nengo.Network() as net:
        inp = nengo.Node([0, 0, 0])
        out = nengo.Node(size_in=3)
        nengo.Connection(inp, out, synapse=None)
        p = nengo.Probe(out)

        # check that sliced input nodes are fed properly as well
        inp2 = nengo.Node([0, 0, 0])
        out2 = nengo.Node(size_in=2)
        nengo.Connection(inp2[:2], out2, synapse=None)
        p2 = nengo.Probe(out2)

    with Simulator(net, minibatch_size=minibatch_size) as sim:
        val = np.random.randn(minibatch_size, 50, 3)
        sim.run_steps(50, data={inp: val, inp2: val})
        assert np.allclose(sim.data[p], val)
        assert np.allclose(sim.data[p2], val[..., :2])

        # error for wrong minibatch size
        with pytest.raises(ValidationError):
            sim.run_steps(10, data={inp: np.zeros((minibatch_size + 1, 10, 3))})
        # error for wrong number of steps
        with pytest.raises(ValidationError):
            sim.run_steps(10, data={inp: np.zeros((minibatch_size, 11, 3))})

        # check that deprecated input_feeds argument also works
        sim.soft_reset(include_probes=True)
        with pytest.warns(DeprecationWarning):
            sim.run_steps(50, input_feeds={inp: val, inp2: val})
        assert np.allclose(sim.data[p], val)
        assert np.allclose(sim.data[p2], val[..., :2])


@pytest.mark.parametrize("neurons", (True, False))
@pytest.mark.training
def test_train_ff(Simulator, neurons, seed):
    minibatch_size = 4
    n_hidden = 20

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1])
        net.config[nengo.Ensemble].bias = nengo.dists.Choice([0])
        net.config[nengo.Connection].synapse = None

        # note: we have these weird input setup just so that we can test
        # training with two distinct inputs
        inp_a = nengo.Node([0])
        inp_b = nengo.Node([0])
        inp = nengo.Node(size_in=2)
        nengo.Connection(inp_a, inp[0])
        nengo.Connection(inp_b, inp[1])

        ens = nengo.Ensemble(
            n_hidden + 1, n_hidden, neuron_type=nengo.Sigmoid(tau_ref=1)
        )
        out = nengo.Ensemble(1, 1, neuron_type=nengo.Sigmoid(tau_ref=1))
        nengo.Connection(inp, ens.neurons if neurons else ens, transform=dists.Glorot())
        nengo.Connection(
            ens.neurons if neurons else ens, out.neurons, transform=dists.Glorot()
        )

        p = nengo.Probe(out.neurons)

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=1, seed=seed
    ) as sim:
        x = np.asarray([[[0, 0]], [[0, 1]], [[1, 0]], [[1, 1]]])
        y = np.asarray([[[0.1]], [[0.9]], [[0.9]], [[0.1]]])

        sim.train(
            {inp_a: x[..., [0]], inp_b: x[..., [1]], p: y},
            tf_compat.train.AdamOptimizer(0.01),
            n_epochs=500,
        )

        sim.check_gradients(atol=5e-5)

        sim.step(data={inp_a: x[..., [0]], inp_b: x[..., [1]]})

        assert np.allclose(sim.data[p], y, atol=2e-3)


@pytest.mark.parametrize("truncation", (None, 5))
@pytest.mark.training
def test_train_recurrent(Simulator, truncation, seed):
    batch_size = 100
    minibatch_size = 100
    n_hidden = 30
    n_steps = 10

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([0])
        ens = nengo.Ensemble(
            n_hidden,
            1,
            neuron_type=nengo.RectifiedLinear(),
            gain=np.ones(n_hidden),
            bias=np.linspace(-1, 1, n_hidden),
        )
        out = nengo.Node(size_in=1)

        nengo.Connection(inp, ens, synapse=None)
        nengo.Connection(ens, ens, synapse=0)
        nengo.Connection(ens, out, synapse=None)

        p = nengo.Probe(out)

    with Simulator(net, minibatch_size=minibatch_size, seed=seed) as sim:
        x = np.outer(np.linspace(0, 1, batch_size), np.ones(n_steps))[:, :, None]
        y = np.outer(np.linspace(0, 1, batch_size), np.linspace(0, 1, n_steps))[
            :, :, None
        ]

        sim.train(
            {inp: x, p: y},
            tf_compat.train.RMSPropOptimizer(1e-3),
            n_epochs=200,
            truncation=truncation,
        )

        sim.check_gradients(sim.tensor_graph.build_outputs({p: mse})[0][p])

        sim.run_steps(n_steps, data={inp: x[:minibatch_size]})

    assert np.sqrt(np.mean((sim.data[p] - y[:minibatch_size]) ** 2)) < (
        0.1 if truncation else 0.055
    )


@pytest.mark.parametrize("unroll", (1, 2))
@pytest.mark.training
def test_train_objective(Simulator, unroll, seed):
    minibatch_size = 1
    n_hidden = 20
    n_steps = 10

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([1])

        ens = nengo.Ensemble(n_hidden, 1, neuron_type=nengo.RectifiedLinear())
        nengo.Connection(inp, ens, synapse=0.01)
        p = nengo.Probe(ens)

        ens2 = nengo.Ensemble(n_hidden, 1, neuron_type=nengo.RectifiedLinear())
        nengo.Connection(inp, ens2, synapse=0.01)
        p2 = nengo.Probe(ens2)

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=unroll, seed=seed
    ) as sim:
        x = np.ones((minibatch_size, n_steps, 1))
        y = np.zeros((minibatch_size, n_steps, 1))
        z = np.zeros((minibatch_size, n_steps, 1)) + 0.1

        def obj(output, target):
            return tf.reduce_mean(
                input_tensor=(output[:, -1] - 0.5 - target[:, -1]) ** 2
            )

        sim.train(
            {inp: x, p: y, p2: z},
            tf_compat.train.MomentumOptimizer(1e-2, 0.9),
            n_epochs=200,
            objective={p: obj, p2: obj},
        )

        sim.check_gradients([p, p2])

        sim.run_steps(n_steps, data={inp: x})

        assert np.allclose(sim.data[p][:, -1], y[:, -1] + 0.5, atol=1e-3)
        assert np.allclose(sim.data[p2][:, -1], z[:, -1] + 0.5, atol=1e-3)


@pytest.mark.training
def test_train_sparse(Simulator, seed):
    minibatch_size = 4
    n_hidden = 20

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1])
        net.config[nengo.Ensemble].bias = nengo.dists.Choice([0])
        net.config[nengo.Ensemble].neuron_type = nengo.RectifiedLinear()
        net.config[nengo.Connection].synapse = None

        inp = nengo.Node([0, 0, 0, 0, 0])
        ens = nengo.Ensemble(n_hidden, 1)
        out = nengo.Node(size_in=2)
        nengo.Connection(inp[[0, 2, 3]], ens.neurons, transform=dists.Glorot())
        nengo.Connection(ens.neurons, out, transform=dists.Glorot())

        p = nengo.Probe(out)

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=1, seed=seed
    ) as sim:
        x = np.asarray(
            [[[0, 0, 0, 0, 0]], [[0, 0, 1, 0, 0]], [[1, 0, 0, 0, 0]], [[1, 0, 1, 0, 0]]]
        )
        y = np.asarray([[[0, 1]], [[1, 0]], [[1, 0]], [[0, 1]]])

        sim.train(
            {inp: x, p: y},
            tf_compat.train.MomentumOptimizer(0.1, 0.9, use_nesterov=True),
            n_epochs=500,
        )

        sim.step(data={inp: x})

        assert np.allclose(sim.data[p], y, atol=1e-3)


@pytest.mark.training
def test_train_errors(Simulator):
    net, a, p = dummies.linear_net()

    n_steps = 20
    with Simulator(net) as sim:
        # error for mismatched n_steps
        with pytest.raises(ValidationError):
            sim.train(
                {a: np.ones((1, n_steps + 1, 1)), p: np.ones((1, n_steps, 1))}, None
            )

        # error for mismatched batch size
        with pytest.raises(ValidationError):
            sim.train({a: np.ones((2, n_steps, 1)), p: np.ones((1, n_steps, 1))}, None)

        # error for mismatched n_steps (in targets)
        with pytest.raises(ValidationError):
            sim.train(
                {a: np.ones((1, n_steps, 1)), p: np.ones((1, n_steps + 1, 1))}, None
            )

        # error for mismatched batch size (in targets)
        with pytest.raises(ValidationError):
            sim.train({a: np.ones((1, n_steps, 1)), p: np.ones((2, n_steps, 1))}, None)

        # error when not specifying objective as a dict
        with pytest.raises(ValidationError):
            sim.train(
                {a: np.ones((1, n_steps, 1)), p: np.ones((1, n_steps, 1))},
                None,
                objective=mse,
            )

        # error when using the old nengo-dl<2.0 inputs/targets style
        with pytest.raises(ValidationError):
            sim.train(
                {a: np.ones((1, n_steps, 1))}, {p: np.ones((1, n_steps, 1))}, None
            )

        # deprecation warning when using "mse" string
        with pytest.warns(DeprecationWarning):
            sim.train(
                {p: np.ones((1, n_steps, 1))},
                tf_compat.train.GradientDescentOptimizer(0),
                objective={p: "mse"},
            )

        # must specify objective if no data
        with pytest.raises(ValidationError):
            sim.train(5, tf_compat.train.GradientDescentOptimizer(0.1))

    # error when calling train after closing
    with pytest.raises(SimulatorClosed):
        sim.train({None: np.zeros((1, 1))}, None)

    with Simulator(net, unroll_simulation=2) as sim:
        # error when data n_steps does not match unroll
        with pytest.raises(ValidationError):
            sim.train({a: np.ones((1, 1, 1)), p: np.ones((1, 1, 1))}, None)

        # error when n_steps does not evenly divide by truncation
        with pytest.raises(ValidationError):
            sim.train(
                {a: np.ones((1, 4, 1)), p: np.ones((1, 4, 1))}, None, truncation=3
            )


@pytest.mark.training
def test_train_no_data(Simulator):
    net, _, p = dummies.linear_net()

    with Simulator(net) as sim:
        sim.train(
            5,
            tf_compat.train.GradientDescentOptimizer(0.1),
            objective={p: lambda x: abs(2.0 - x)},
            n_epochs=10,
        )
        sim.step()
        assert np.allclose(sim.data[p], 2)


def test_loss(Simulator):
    with nengo.Network() as net:
        inp = nengo.Node([1])
        ens = nengo.Ensemble(30, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    n_steps = 20
    with Simulator(net) as sim:
        sim.run_steps(n_steps)
        data = sim.data[p]

        # check default mse objective
        assert np.allclose(
            sim.loss({inp: np.ones((4, n_steps, 1)), p: np.zeros((4, n_steps, 1))}),
            np.mean(data ** 2),
        )

        # check custom objective
        assert np.allclose(
            sim.loss(
                {inp: np.ones((4, n_steps, 1)), p: np.zeros((4, n_steps, 1))},
                {p: lambda x, y: tf.constant(2.0)},
            ),
            2,
        )

        # error for mismatched n_steps
        with pytest.raises(ValidationError):
            sim.loss({inp: np.ones((1, n_steps + 1, 1)), p: np.ones((1, n_steps, 1))})

        # error for mismatched batch size
        with pytest.raises(ValidationError):
            sim.loss({inp: np.ones((2, n_steps, 1)), p: np.ones((1, n_steps, 1))})

        # error for mismatched n_steps (in targets)
        with pytest.raises(ValidationError):
            sim.loss({inp: np.ones((1, n_steps, 1)), p: np.ones((1, n_steps + 1, 1))})

        # error for mismatched batch size (in targets)
        with pytest.raises(ValidationError):
            sim.loss({inp: np.ones((1, n_steps, 1)), p: np.ones((2, n_steps, 1))})

        # error when not specifying objective as a dict
        with pytest.raises(ValidationError):
            sim.loss({inp: np.ones((1, n_steps, 1)), p: np.ones((1, n_steps, 1))}, mse)

        # deprecation warning when using "mse" string
        with pytest.warns(DeprecationWarning):
            sim.loss({p: np.ones((1, n_steps, 1))}, {p: "mse"})

        # must specify objective if no data
        with pytest.raises(ValidationError):
            sim.loss(5)

    # error when calling loss after close
    with pytest.raises(SimulatorClosed):
        sim.loss({None: np.zeros((1, 1))})

    with Simulator(net, unroll_simulation=2) as sim:
        # error when data n_steps does not match unroll
        with pytest.raises(ValidationError):
            sim.loss({inp: np.ones((1, 1, 1)), p: np.ones((1, 1, 1))})


def test_generate_inputs(Simulator, seed):
    with nengo.Network() as net:
        proc = nengo.processes.WhiteNoise(seed=seed)
        inp = [
            nengo.Node([1]),
            nengo.Node(np.sin),
            nengo.Node(proc),
            nengo.Node([2]),
            nengo.Node(nengo.processes.WhiteNoise()),
        ]

        p = [nengo.Probe(x) for x in inp]

    with Simulator(net, minibatch_size=2, unroll_simulation=3) as sim:
        feed = sim._generate_inputs({inp[0]: np.zeros((2, 3, 1))}, 3)

        ph = [sim.tensor_graph.input_ph[x] for x in inp]

        assert len(sim.tensor_graph.invariant_inputs) == len(inp)
        assert len(feed) == len(inp)

        sim.reset()
        sim.run_steps(3, data={inp[0]: np.zeros((2, 3, 1))})

        vals = [
            np.zeros((3, 1, 2)),
            np.tile(np.sin(sim.trange())[:, None, None], (1, 1, 2)),
            np.tile(proc.run_steps(3)[:, :, None], (1, 1, 2)),
            np.ones((3, 1, 2)) * 2,
        ]
        for i, x in enumerate(vals):
            assert np.allclose(feed[ph[i]], x)
            assert np.allclose(sim.data[p[i]], x.transpose(2, 0, 1))

        # check that unseeded process was different in each minibatch item
        assert not np.allclose(feed[ph[-1]][..., 0], feed[ph[-1]][..., 1])


@pytest.mark.training
def test_save_load_params(Simulator, tmpdir):
    with nengo.Network(seed=0) as net:
        inp = nengo.Node([0])
        out = nengo.Node(size_in=1)
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        nengo.Connection(ens, out)

        configure_settings(trainable=None)
        net.config[ens].trainable = False

    with Simulator(net) as sim:
        weights_var = [
            x[0]
            for x in sim.tensor_graph.base_vars.values()
            if x[0].get_shape() == (1, 10)
        ][0]
        enc_var = sim.tensor_graph.base_vars[
            sim.tensor_graph.signals[sim.model.sig[ens]["encoders"]].key
        ][0]
        weights0, enc0 = sim.sess.run([weights_var, enc_var])
        sim.save_params(os.path.join(str(tmpdir), "train"))
        sim.save_params(os.path.join(str(tmpdir), "local"), include_local=True)

    with pytest.raises(SimulatorClosed):
        sim.save_params(None)
    with pytest.raises(SimulatorClosed):
        sim.load_params(None)

    with nengo.Network(seed=1) as net2:
        inp = nengo.Node([0])
        out = nengo.Node(size_in=1)
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        nengo.Connection(ens, out)

        configure_settings(trainable=None)
        net2.config[ens].trainable = False

    with Simulator(net2) as sim:
        weights_var = [
            x[0]
            for x in sim.tensor_graph.base_vars.values()
            if x[0].get_shape() == (1, 10)
        ][0]
        enc_var = sim.tensor_graph.base_vars[
            sim.tensor_graph.signals[sim.model.sig[ens]["encoders"]].key
        ][0]
        weights1, enc1 = sim.sess.run([weights_var, enc_var])
        assert not np.allclose(weights0, weights1)
        assert not np.allclose(enc0, enc1)

        sim.load_params(os.path.join(str(tmpdir), "train"))

        weights2, enc2 = sim.sess.run([weights_var, enc_var])
        assert np.allclose(weights0, weights2)
        assert not np.allclose(enc0, enc2)

        sim.load_params(os.path.join(str(tmpdir), "local"), include_local=True)

        weights3, enc3 = sim.sess.run([weights_var, enc_var])
        assert np.allclose(weights0, weights3)
        assert np.allclose(enc0, enc3)


def test_model_passing(Simulator, seed):
    # make sure that passing a built model to the Simulator works properly

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([1])
        ens = nengo.Ensemble(20, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    model = nengo.builder.Model()
    model.build(net)

    ops = model.operators.copy()

    with nengo.Simulator(None, model=model, optimize=False) as sim:
        sim.run_steps(10)

    assert ops == model.operators
    canonical = sim.data[p]

    with Simulator(None, model=model) as sim:
        sim.run_steps(10)

    assert ops == model.operators
    assert np.allclose(sim.data[p], canonical)

    # make sure that passing the same model to Simulator twice works
    with Simulator(None, model=model) as sim:
        sim.run_steps(10)
        assert ops == model.operators

    assert ops == model.operators
    assert np.allclose(sim.data[p], canonical)

    # make sure that passing that model back to the reference simulator works
    with nengo.Simulator(None, model=model, optimize=False) as sim:
        sim.run_steps(10)

    assert ops == model.operators
    assert np.allclose(sim.data[p], canonical)


@pytest.mark.parametrize("device", ["/cpu:0", "/gpu:0", None])
def test_devices(Simulator, device, seed, caplog):
    if device == "/gpu:0" and not tf_gpu_installed:
        pytest.skip("This test requires tensorflow-gpu")

    caplog.set_level(logging.INFO)

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([0])
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    with nengo.Simulator(net) as sim:
        sim.run_steps(50)
        canonical = sim.data[p]

    with Simulator(net, device=device) as sim:
        sim.run_steps(50)
        assert np.allclose(sim.data[p], canonical)

        if device == "/cpu:0":
            assert "Running on CPU" in caplog.text
        elif device == "/gpu:0":
            assert "Running on GPU" in caplog.text
        elif tf_gpu_installed:
            assert "Running on CPU/GPU" in caplog.text
        else:
            # device is None but gpu not installed
            assert "Running on CPU" in caplog.text


def test_side_effects(Simulator):
    class MyFunc:
        x = 0

        def __call__(self, t):
            self.x += 1

    func = MyFunc()

    with nengo.Network() as net:
        nengo.Node(output=func)

    with Simulator(net, unroll_simulation=1) as sim:
        sim.run_steps(10)

    assert func.x == 11


@pytest.mark.training
def test_tensorboard(Simulator, tmpdir):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1, neuron_type=nengo.LIFRate())
        c = nengo.Connection(a, b)
        p = nengo.Probe(b)
        p2 = nengo.Probe(c)

    with Simulator(net, tensorboard=str(tmpdir)):
        assert os.path.exists("%s/run_0" % tmpdir)

    # check that the run incrementing works properly
    with Simulator(net, tensorboard=str(tmpdir)):
        assert os.path.exists("%s/run_1" % tmpdir)

    # check that training summaries are output properly
    with Simulator(net, tensorboard=str(tmpdir)) as sim:
        with tf.device("/cpu:0"):
            summ = tf_compat.summary.scalar("step_var", sim.training_step)

        def loss(x):
            # uses a variable to test that variables from summaries get
            # initialized correctly
            return tf_compat.get_variable(
                "one",
                initializer=tf_compat.initializers.constant(1.0),
                dtype=x.dtype,
                shape=(),
                use_resource=False,
            )

        sim.train(
            {
                a: np.zeros((1, 10, 1)),
                p: np.zeros((1, 10, 1)),
                p2: np.zeros((1, 10, 1)),
            },
            tf_compat.train.GradientDescentOptimizer(0.0),
            objective={p: loss, p2: mse},
            summaries=["loss", b, b.neurons, c, summ],
            n_epochs=3,
        )

    event_file = os.path.join(str(tmpdir), "run_2", os.listdir("%s/run_2" % tmpdir)[0])

    assert os.path.exists(event_file)

    for i, event in enumerate(tf_compat.train.summary_iterator(event_file)):
        if i < 3:
            # metadata stuff
            continue

        assert event.step == i - 2
        tags = [s.tag for s in event.summary.value]
        assert len(tags) == 7
        assert "loss/loss" in tags[0]
        assert "loss/Probe_None_loss" in tags[1]
        assert "loss/Probe_None_loss" in tags[2]
        assert tags[3] == "Ensemble_None_encoders"
        assert tags[4] == "Ensemble.neurons_None_bias"
        assert tags[5] == "Connection_None_weights"
        assert tags[6] == "step_var"

    assert i == 5  # pylint: disable=undefined-loop-variable

    # check for warning if user requests summaries with tensorboard=None
    with pytest.warns(UserWarning):
        with Simulator(net, tensorboard=None) as sim:
            sim.train(
                {a: np.zeros((1, 10, 1)), p: np.zeros((1, 10, 1))},
                tf_compat.train.GradientDescentOptimizer(0.0),
                summaries=["loss"],
            )

    # check for error on invalid object
    with pytest.raises(SimulationError):
        with Simulator(net, tensorboard=str(tmpdir)) as sim:
            sim.train(
                {a: np.zeros((1, 10, 1)), p: np.zeros((1, 10, 1))},
                tf_compat.train.GradientDescentOptimizer(0.0),
                summaries=[a],
            )

    # check that nonexistent dir gets created
    with Simulator(net, tensorboard=str(tmpdir.join("new"))):
        assert os.path.exists(str(tmpdir.join("new")))


@pytest.mark.parametrize("mode", ("run", "train"))
@pytest.mark.parametrize("outfile", (None, "tmp.txt"))
@pytest.mark.training
def test_profile(Simulator, mode, outfile):
    net, a, p = dummies.linear_net()

    if outfile is None:
        filename = "nengo_dl_profile.json"
    else:
        filename = outfile
    if os.path.exists(filename):
        os.remove(filename)

    with Simulator(net) as sim:
        if outfile is None:
            prof = True
        else:
            prof = outfile

        if mode == "run":
            sim.run_steps(5, profile=prof)
        else:
            sim.train(
                {a: np.zeros((1, 5, 1)), p: np.zeros((1, 5, 1))},
                tf_compat.train.GradientDescentOptimizer(1),
                profile=prof,
            )

        assert os.path.exists(filename)
        os.remove(filename)


def test_dt_readonly(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        nengo.Probe(a)

    with Simulator(net) as sim:
        with pytest.raises(ReadonlyError):
            sim.dt = 1


def test_probe_data():
    sim = dummies.Simulator()
    a = dummies.Probe(add_to_container=False)
    b = dummies.Probe(add_to_container=False)
    sim.model.params = OrderedDict(
        {
            a: [np.zeros((5, 1, 3)), np.ones((5, 1, 3))],
            b: [np.ones((1, 1, 3)), np.zeros((1, 1, 3))],
        }
    )
    sim.model.probes = (a, b)
    data = SimulationData(sim, True)
    assert data[a].shape == (5, 2, 3)
    assert np.all(data[a][:, 0] == 0)
    assert np.all(data[a][:, 1] == 1)

    data.minibatched = False
    assert data[b].shape == (2, 3)
    assert np.all(data[b][1] == 0)
    assert np.all(data[b][0] == 1)


@pytest.mark.parametrize(
    "pre_val", [1, lambda t: 1, nengo.processes.Piecewise({0: [1]})]
)
@pytest.mark.parametrize(
    "post_val", [2, lambda t: 2, nengo.processes.Piecewise({0: [2]})]
)
def test_node_output_change(Simulator, pre_val, post_val, seed):
    with nengo.Network(seed=seed) as net:
        inp = nengo.Node(pre_val)
        p = nengo.Probe(inp)

    with Simulator(net, unroll_simulation=1) as sim:
        sim.step()
        inp.output = post_val
        sim.step()

        assert np.allclose(sim.data[p], 1.0)

        sim.reset()
        sim.step()

        assert np.allclose(sim.data[p], 1.0)


@pytest.mark.training
def test_check_gradients_error(Simulator):
    # check_gradients detects nans in gradient
    with nengo.Network() as net:
        x = nengo.Node([0])
        y = tensor_layer(x, lambda x: 1 / x)
        nengo.Probe(y)

    with Simulator(net) as sim:
        with pytest.raises(SimulationError):
            sim.check_gradients()

    # check_gradients detects errors in gradient (in this case caused by the
    # fact that nengo.Triangle doesn't have a TensorFlow implementation)
    with nengo.Network() as net:
        x = nengo.Node([0])
        nengo.Probe(x, synapse=nengo.Triangle(0.1))

    with Simulator(net) as sim:
        with pytest.raises(SimulationError):
            sim.check_gradients()


def test_check_data(Simulator):
    with nengo.Network() as net:
        inpa = nengo.Node([0, 0])
        inpb = nengo.Node([0])
        n = nengo.Node(size_in=2)
        pa = nengo.Probe(inpa)
        pb = nengo.Probe(inpb)
        e = nengo.Ensemble(10, 1)

    with nengo.Network():
        inp2 = nengo.Node([0, 0])
        p2 = nengo.Probe(inp2)

    with Simulator(net, minibatch_size=3) as sim:
        zeros1 = np.zeros((3, 1, 1))
        zeros2 = np.zeros((3, 1, 2))

        # make sure that valid inputs pass
        sim._check_data({inpa: zeros2, inpb: zeros1})
        sim._check_data({pa: zeros2, pb: zeros1})

        # invalid input objects
        with pytest.raises(ValidationError):
            sim._check_data({inp2: zeros2})
        with pytest.raises(ValidationError):
            sim._check_data({n: zeros2})

        # invalid target object
        with pytest.raises(ValidationError):
            sim._check_data({p2: zeros2})

        # mismatched input data
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2, inpb: np.zeros((2, 1, 1))})
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2, inpb: np.zeros((1, 2, 1))})

        # mismatched target data
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros2, pb: np.zeros((2, 1, 1))})
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros2, pb: np.zeros((1, 2, 1))})

        # data that doesn't match explicit target
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2}, n_batch=2)
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2}, n_steps=2)
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros2}, n_batch=2)
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros2}, n_steps=2)

        # data with wrong object dimensionality
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros1})
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros1})

        # data with batch size < minibatch_size
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2[[0]]})
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros2[[0]]})

        # data with incorrect rank
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2[0]})
        with pytest.raises(ValidationError):
            sim._check_data({pa: zeros2[0]})

        # invalid object type in data
        with pytest.raises(ValidationError):
            sim._check_data({inpa: zeros2, e: zeros1})


def test_matching_node_out(Simulator):
    # make sure that nodes with identical outputs are handled correctly (the
    # outputs will have the same hash, so that can cause some errors)

    with nengo.Network() as net:
        a = nengo.Node(output=nengo.processes.WhiteSignal(1, 10), size_out=2)
        b = nengo.Node(output=nengo.processes.WhiteSignal(1, 10), size_out=2)
        c = nengo.Node(output=nengo.processes.WhiteSignal(1, 10), size_out=3)

        p_a = nengo.Probe(a)
        p_b = nengo.Probe(b)
        p_c = nengo.Probe(c)

    with Simulator(net) as sim:
        sim.run_steps(10)

        assert sim.data[p_a].shape == (10, 2)
        assert sim.data[p_b].shape == (10, 2)
        assert not np.allclose(sim.data[p_a], sim.data[p_b])
        assert sim.data[p_c].shape == (10, 3)


def test_probe_no_data(Simulator):
    with nengo.Network() as net:
        u = nengo.Node([0])
        p = nengo.Probe(u)

    with Simulator(net) as sim:
        pass

    assert sim.data[p] == []


@pytest.mark.training
def test_train_state_save(Simulator):
    with nengo.Network() as net:
        u = nengo.Node([1])
        o = nengo.Node(size_in=1)
        nengo.Connection(u, o)
        p = nengo.Probe(u, synapse=0.1)

    with Simulator(net) as sim:
        sim.run_steps(20)

    with Simulator(net) as sim2:
        sim2.run_steps(10)

        sim2.train(
            {u: np.ones((4, 10, 1)), p: np.ones((4, 10, 1))},
            optimizer=tf_compat.train.GradientDescentOptimizer(0),
        )

        sim2.loss({u: np.ones((4, 10, 1)), p: np.ones((4, 10, 1))})

        sim2.run_steps(10)

    assert np.allclose(sim.data[p], sim2.data[p])


def test_simulation_data(Simulator, seed):
    rng = np.random.RandomState(seed)
    N = 17
    d = 2

    gain = rng.uniform(low=1, high=2, size=N)
    bias = rng.uniform(low=0.2, high=1, size=N)
    enc = rng.uniform(-1, 1, size=(N, d))
    enc /= np.linalg.norm(enc, axis=1)[:, None]

    with nengo.Network() as net:
        u = nengo.Node([0] * d)
        a = nengo.Ensemble(N, d, gain=gain, bias=bias, encoders=enc, radius=3)
        b = nengo.Ensemble(
            N,
            d,
            gain=gain,
            bias=bias,
            encoders=enc * 2,
            radius=3,
            normalize_encoders=False,
        )
        c = nengo.Ensemble(N, d, neuron_type=nengo.Direct())
        conn0 = nengo.Connection(u, a)
        conn = nengo.Connection(
            a.neurons,
            b,
            transform=rng.uniform(-1, 1, size=(d, N)),
            learning_rule_type=nengo.PES(),
        )

        p = nengo.Probe(b)

    with Simulator(net) as sim:
        # check dict functions
        assert u in sim.data
        assert a in sim.data
        assert b in sim.data
        assert c in sim.data
        assert conn in sim.data
        assert conn0 in sim.data
        assert p in sim.data
        assert net in sim.data
        assert conn.learning_rule in sim.data
        assert len(sim.data) == 10  # 9 objects + probe connection
        for k, k2 in zip(sim.data, sim.data.keys()):
            assert k is k2

        # check gain/bias
        assert np.allclose(gain, sim.data[a].gain)
        assert np.allclose(bias, sim.data[a].bias)

        # check max_rates/intercepts
        max_rates, intercepts = a.neuron_type.max_rates_intercepts(gain, bias)
        assert np.allclose(max_rates, sim.data[a].max_rates)
        assert np.allclose(intercepts, sim.data[a].intercepts)

        # check encoders/scaled_encoders
        assert np.allclose(enc, sim.data[a].encoders)
        assert np.allclose(enc * gain[:, None] / a.radius, sim.data[a].scaled_encoders)

        # make sure that the inferences still work with non-normalized encoders
        assert np.allclose(enc * 2, sim.data[b].encoders)
        assert np.allclose(gain, sim.data[b].gain)
        assert np.allclose(bias, sim.data[b].bias)
        assert np.allclose(
            enc * 2 * gain[:, None] / b.radius, sim.data[b].scaled_encoders
        )

        # check connection weights
        transform = conn.transform
        if LooseVersion(nengo.__version__) > "2.8.0":
            transform = transform.init
        assert np.allclose(transform, sim.data[conn].weights)

        # check that batch dimension eliminated
        assert sim.data[conn].weights.shape == (d, N)

        # check that values can be updated live
        sig = sim.model.sig[a]["encoders"]
        tensor_sig = sim.tensor_graph.signals[sig]
        base = sim.tensor_graph.base_vars[tensor_sig.key][0]
        op = tf_compat.assign(base, tf.ones_like(base))
        sim.sess.run(op)

        assert np.allclose(sim.data[a].scaled_encoders, 1)
        assert np.allclose(sim.data[a].gain, np.sqrt(2) * a.radius)
        assert np.allclose(sim.data[a].encoders, 1 / np.sqrt(2))

        # check that direct mode ensemble parameters are handled correctly
        assert np.allclose(sim.data[c].encoders, np.eye(d))
        assert sim.data[c].gain is None
        assert sim.data[c].bias is None

    # check minibatch dimension moved to front
    with Simulator(net, minibatch_size=3) as sim:
        assert sim.data[conn].weights.shape == (3, d, N)

    # reverts back to init after simulator close, and warns
    with pytest.warns(UserWarning):
        assert np.allclose(sim.data[a].encoders, enc)

    with pytest.raises(ValidationError):
        _ = sim.data[nengo.Ensemble(10, 1, add_to_container=False)]


@pytest.mark.training
def test_learning_rate_schedule(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1, neuron_type=nengo.LIFRate())
        nengo.Connection(a, b)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        vals = [1.0, 0.1, 0.001]
        with tf.device("/cpu:0"):
            l_rate = tf_compat.train.piecewise_constant(
                sim.training_step,
                [tf.constant(4, dtype=tf.int64), tf.constant(9, dtype=tf.int64)],
                vals,
            )
        opt = tf_compat.train.GradientDescentOptimizer(l_rate)

        for i in range(3):
            assert np.allclose(sim.sess.run(l_rate), vals[i])
            sim.train(
                {a: np.zeros((1, 10, 1)), p: np.zeros((1, 10, 1))}, opt, n_epochs=5
            )


@pytest.mark.training
def test_multiple_objective(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        net.config[nengo.Connection].synapse = None

        a = nengo.Node([0])

        # note: b is configured this way so that the output, and therefore
        # loss, will be equal to the input, so we can control it easily
        b = nengo.Ensemble(
            100,
            1,
            neuron_type=nengo.RectifiedLinear(),
            gain=nengo.dists.Choice([1]),
            bias=nengo.dists.Choice([0]),
            encoders=nengo.dists.Choice([[1]]),
        )

        c = nengo.Ensemble(10, 1)

        nengo.Connection(a, b)
        nengo.Connection(a, c)

        p_b = nengo.Probe(b)
        p_c = nengo.Probe(c)

    with Simulator(net, unroll_simulation=1) as sim:
        data = {
            a: np.ones((10, 1, 1)),
            p_b: np.zeros((10, 1, 1)),
            p_c: np.zeros((10, 1, 1)),
        }
        objective = {p_b: lambda x, y: x, p_c: lambda x, y: x * 0}

        loss = sim.loss(data, objective)
        assert np.allclose(loss, 1, atol=1e-3)

        b_bias = np.copy(sim.data[b].bias)
        c_bias = sim.data[c].bias
        sim.train(
            data,
            tf_compat.train.GradientDescentOptimizer(1.0),
            objective=objective,
            n_epochs=10,
        )
        assert np.allclose(sim.data[c].bias, c_bias)
        assert not np.allclose(sim.data[b].bias, b_bias)


def test_get_nengo_params(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        a = nengo.Ensemble(12, 3, label="a")
        b = nengo.Ensemble(10, 4, label="b", radius=2)
        n = nengo.Node([1])
        c = nengo.Connection(a.neurons[:5], b[:2], transform=Uniform(-1, 1), label="c")
        d = nengo.Connection(
            a,
            b.neurons,
            function=lambda x: np.ones(5),
            transform=Uniform(-1, 1),
            label="d",
        )
        e = nengo.Connection(n, b, transform=Uniform(-1, 1), label="e")
        f = nengo.Ensemble(5, 1, label="a")
        g = nengo.Ensemble(11, 1, neuron_type=nengo.Direct(), label="g")
        if LooseVersion(nengo.__version__) > "2.8.0":
            h = nengo.Connection(
                a.neurons,
                b,
                transform=nengo.Convolution(1, (2, 2, 3), padding="same"),
                label="h",
            )
        p = nengo.Probe(b.neurons)

    with Simulator(net, seed=seed) as sim:
        # check that we get an error for non-ensemble/connection objects
        with pytest.raises(ValueError):
            sim.get_nengo_params(n)

        # check that we get an error for duplicate labels
        with pytest.raises(ValueError):
            sim.get_nengo_params([a, f], as_dict=True)

        # check that we get an error for direct ensembles
        with pytest.raises(ValueError):
            sim.get_nengo_params(g)

        # check that single objects are returned as single dicts
        params = sim.get_nengo_params(d)
        assert params["transform"] == 1

        fetches = [a.neurons, b, c, d, e]
        if LooseVersion(nengo.__version__) > "2.8.0":
            fetches += [h]

        params = sim.get_nengo_params(fetches, as_dict=True)
        sim.run_steps(100)

    with nengo.Network(seed=seed + 1) as net:
        a2 = nengo.Ensemble(12, 3, **params["a"])
        b2 = nengo.Ensemble(10, 4, radius=2, **params["b"])
        n2 = nengo.Node([1])
        nengo.Connection(a2.neurons[:5], b2[:2], **params["c"])
        nengo.Connection(a2, b2.neurons, **params["d"])
        nengo.Connection(n2, b2, **params["e"])
        if LooseVersion(nengo.__version__) > "2.8.0":
            nengo.Connection(a2.neurons, b2, **params["h"])
        p2 = nengo.Probe(b2.neurons)

    with Simulator(net, seed=seed) as sim2:
        sim2.run_steps(100)

        assert np.allclose(sim.data[p], sim2.data[p2])


@pytest.mark.parametrize("progress", (True, False))
def test_progress_bar(Simulator, progress):
    net, _, p = dummies.linear_net()

    # note: ideally we would capture the stdout and check that output is
    # actually being controlled. but the pytest capturing doesn't work,
    # because it's being printed in a different thread (I think). so we just
    # check that the parameter works without error
    with Simulator(net, progress_bar=progress) as sim:
        sim.run_steps(10, progress_bar=progress)
        sim.loss(10, {p: lambda x: x}, progress_bar=progress)

        if not sim.tensor_graph.inference_only:
            sim.train(
                10,
                tf_compat.train.GradientDescentOptimizer(0),
                objective={p: lambda x: x},
                progress_bar=progress,
            )


@pytest.mark.training
def test_extra_feeds(Simulator):
    # set up a tensornode that will fail unless a value is fed in for the
    # placeholder
    class NodeFunc:
        def pre_build(self, *_):
            self.ph = tf_compat.placeholder_with_default(False, ())

        def __call__(self, t, x):
            with tf.device("/cpu:0"):
                check = tf.Assert(self.ph, [t])
            with tf.control_dependencies([check]):
                y = tf.identity(x)
            return y

    with nengo.Network() as net:
        a = nengo.Node([0])
        b = TensorNode(NodeFunc(), size_in=1, size_out=1)
        nengo.Connection(a, b)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        with pytest.raises(tf.errors.InvalidArgumentError):
            sim.run_steps(10)
        sim.run_steps(10, extra_feeds={b.tensor_func.ph: True})

        data = {a: np.zeros((1, 10, 1)), p: np.zeros((1, 10, 1))}
        with pytest.raises(tf.errors.InvalidArgumentError):
            sim.train(data, tf_compat.train.GradientDescentOptimizer(1))
        sim.train(
            data,
            tf_compat.train.GradientDescentOptimizer(1),
            extra_feeds={b.tensor_func.ph: True},
        )

        with pytest.raises(tf.errors.InvalidArgumentError):
            sim.loss(data)
        sim.loss(data, extra_feeds={b.tensor_func.ph: True})


@pytest.mark.parametrize("mixed", (False, True))
@pytest.mark.training
def test_direct_grads(Simulator, mixed):
    net, a, p = dummies.linear_net()

    if mixed:
        with net:
            c = nengo.Ensemble(
                1,
                1,
                neuron_type=nengo.RectifiedLinear(),
                gain=np.ones(1),
                bias=np.ones(1) * 1e-6,
            )
            net.config[c.neurons].trainable = False
            nengo.Connection(a, c.neurons, synapse=None)
            p2 = nengo.Probe(c.neurons)

    with Simulator(net, minibatch_size=1) as sim:
        n_steps = 10
        opt = tf_compat.train.GradientDescentOptimizer(0.45)
        for _ in range(10):
            sim.run_steps(n_steps)

            data = {p: (2.0 / n_steps) * (sim.data[p] - 2)}
            obj = {p: None}

            if mixed:
                data[p2] = np.ones((1, n_steps, 1)) * 2
                obj[p2] = mse
                assert np.allclose(sim.data[p], sim.data[p2])

            data[a] = np.ones((1, n_steps, 1))
            sim.train(data, opt, objective=obj)
            sim.soft_reset(include_probes=True)

        sim.run_steps(n_steps)
        assert np.allclose(sim.data[p], 2)
        if mixed:
            assert np.allclose(sim.data[p2], 2)


@pytest.mark.training
def test_non_differentiable(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Node(lambda t, x: x, size_in=1)
        c = nengo.Connection(a, b)
        nengo.Connection(a, b)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        w0 = sim.data[c].weights
        sim.train(
            {a: np.ones((1, 10, 1)), p: np.ones((1, 10, 1))},
            tf_compat.train.GradientDescentOptimizer(100),
        )

        # TODO: find another way to detect non-differentiable elements in graph
        # note: the challenge is that our stateful ops tend to mask the
        # backpropagating None gradients, because the overwritten parts of
        # the variable end up with a gradient of zero (regardless of what the
        # output gradient is)
        assert np.allclose(sim.data[c].weights, w0)


@pytest.mark.parametrize(
    "net",
    (
        nengo.networks.BasalGanglia(4),
        nengo.networks.CircularConvolution(32, 4),
        nengo.networks.Oscillator(0.01, 10, 100),
        nengo.networks.InputGatedMemory(32, 4),
    ),
)
def test_freeze_network(Simulator, net):
    with Simulator(net) as sim:
        sim.run(0.1)
        sim.freeze_params(net)

    with nengo.Simulator(net) as sim2:
        sim2.run(0.1)

    for p in net.all_probes:
        assert np.allclose(sim.data[p], sim2.data[p])


def test_freeze_obj(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        a = nengo.Ensemble(10, 1)
        b = nengo.Node(size_in=1)
        c = nengo.Connection(a, b)
        p = nengo.Probe(b)

    with nengo.Network(seed=seed + 1):
        d = nengo.Ensemble(10, 1)

    with Simulator(net) as sim:
        # check for error with object outside network
        with pytest.raises(ValueError):
            sim.freeze_params(d)

        # check for error with wrong object type
        with pytest.raises(TypeError):
            sim.freeze_params(p)

        sim.freeze_params([a, c])
        sim.run_steps(10)

    with nengo.Simulator(net) as sim2:
        sim2.run_steps(10)

    assert np.allclose(sim.data[p], sim2.data[p])

    # check that we get an error if the simulator is closed
    with pytest.raises(SimulatorClosed):
        sim.freeze_params(net)


@pytest.mark.training
def test_freeze_train(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([1])
        b = nengo.Node(size_in=1)
        nengo.Connection(a, b, transform=1, synapse=None)
        p = nengo.Probe(b)

    with Simulator(net, unroll_simulation=1) as sim:
        sim.step()
        assert np.allclose(sim.data[p], 1)

        sim.train(
            {a: np.ones((1, 1, 1)), p: np.ones((1, 1, 1)) * 2},
            tf_compat.train.GradientDescentOptimizer(0.5),
            n_epochs=10,
        )

        sim.step()
        assert np.allclose(sim.data[p][-1], 2)

        sim.freeze_params(net)

    with nengo.Simulator(net) as sim:
        sim.step()
        assert np.allclose(sim.data[p], 2)


def test_fill_feed(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        p0 = nengo.Probe(a)
        p1 = nengo.Probe(a)

    with Simulator(net) as sim:
        # build an objective with p0 in it, so that it will be added to the
        # graph
        sim.tensor_graph.build_outputs({p0: mse})

        # filling p0 will work fine
        sim._fill_feed(1, data={p0: np.zeros((1, 1, 1))})

        # validation error if filling p1
        with pytest.raises(ValidationError):
            sim._fill_feed(1, data={p1: np.zeros((1, 1, 1))})


@pytest.mark.parametrize("neuron_type", (nengo.SpikingRectifiedLinear(), nengo.LIF()))
def test_inference_only(Simulator, neuron_type, seed):
    with nengo.Network(seed=seed) as net:
        configure_settings(inference_only=False)

        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1, neuron_type=neuron_type)
        c = nengo.Connection(a, b, synapse=None)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        sim.run(0.1)

        assert sim.model.sig[c]["weights"].trainable

    with net:
        configure_settings(inference_only=True)

    with Simulator(net) as sim2:
        sim2.run(0.1)

        # check that inference-only mode produces the same output
        assert np.allclose(sim.data[p], sim2.data[p])

        # check that parameters aren't trainable
        assert not sim2.model.sig[c]["weights"].trainable

        # validation checks (can't do train/gradients in inference-only mode)
        with pytest.raises(ValidationError):
            sim2.train(
                {a: np.zeros((1, 10, 1)), p: np.zeros((1, 10, 1))},
                tf_compat.train.GradientDescentOptimizer(1),
            )

        with pytest.raises(ValidationError):
            sim2.check_gradients()


def test_dtype(Simulator):
    with pytest.warns(DeprecationWarning):
        with Simulator(None, dtype=tf.float32) as sim:
            assert sim.tensor_graph.dtype == tf.float32


@pytest.mark.training
def test_synapse_warning(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1)
        c = nengo.Connection(a, b, synapse=1)
        p = nengo.Probe(b)
        p2 = nengo.Probe(b)

    def does_warn(n_steps=1, as_dict=True):
        with Simulator(net, unroll_simulation=1) as sim:
            with pytest.warns(None) as rec:
                if as_dict:
                    sim.train(
                        {a: np.zeros((1, n_steps, 1)), p: np.zeros((1, n_steps, 1))},
                        tf_compat.train.GradientDescentOptimizer(0),
                    )
                else:
                    sim.train(
                        n_steps,
                        tf_compat.train.GradientDescentOptimizer(0),
                        objective={p: lambda x: x},
                    )
        return any(str(w.message).startswith("Training for one timestep") for w in rec)

    # warning from connection
    assert does_warn()

    # warning from probe
    c.synapse = None
    p.synapse = 1
    assert does_warn()

    # no warning for >1 step training
    assert not does_warn(n_steps=2)

    # no warning from non-target probe
    p.synapse = None
    p2.synapse = 1
    assert not does_warn()

    # warning when explicitly specifying n_steps
    c.synapse = 1
    assert does_warn(as_dict=True)


@pytest.mark.training
def test_concat_hang(Simulator, pytestconfig):
    if (
        "1.11.0" <= LooseVersion(tf.__version__) < "1.14"
        and pytestconfig.getoption("--unroll-simulation") > 1
    ):
        pytest.xfail(
            "There is a bug in TensorFlow that causes this test to hang; see "
            "https://github.com/tensorflow/tensorflow/issues/23383"
        )

    with nengo.Network() as net:
        a = nengo.Node([0])
        x = nengo.Node(size_in=1)
        nengo.Connection(a, x)
        p = nengo.Probe(x)

    with Simulator(net) as sim:
        sim.train(
            {a: np.zeros((1, 5, 1)), p: np.zeros((1, 5, 1))},
            tf_compat.train.GradientDescentOptimizer(1),
        )


def test_run_batch(Simulator, rng):
    counter = 0

    def count(t):
        nonlocal counter
        counter += 1
        return [t]

    with nengo.Network() as net:
        a = nengo.Node(count, size_out=1)
        p = nengo.Probe(a)

    with Simulator(net, minibatch_size=2) as sim:
        # check that run_batch works with empty outputs
        sim.run_batch(10, {})
        assert counter == 10

        # check handling of output tuples
        inp = rng.randn(10, 10, 1)
        out = sim.run_batch({a: inp}, {p: lambda x: (x, x + 1)})[p]
        assert len(out) == 2
        assert isinstance(out, tuple)
        assert out[0].shape == out[1].shape == (5, 2, 10, 1)
        inp = np.reshape(inp, (5, 2, 10, 1))
        assert np.allclose(out[0], inp)
        assert np.allclose(out[1], inp + 1)


def test_tf_seed(Simulator, seed):
    with nengo.Network() as net:
        a = TensorNode(lambda t: tf_uniform((1, 1), dtype=t.dtype))
        p = nengo.Probe(a)

    with Simulator(net, seed=seed) as sim:
        sim.step()

    with Simulator(net, seed=seed) as sim_seed:
        sim_seed.step()

    assert np.allclose(sim.data[p], sim_seed.data[p])

    with Simulator(net, seed=seed + 1) as sim_reset:
        sim_reset.step()

        assert not np.allclose(sim.data[p], sim_reset.data[p])

        sim_reset.reset(seed=seed)

        assert np.allclose(sim.data[p], sim_reset.data[p])


def test_pickle_error(Simulator):
    with pytest.raises(NotImplementedError, match="does not support pickling"):
        pickle.dumps(Simulator(nengo.Network()))



# pylint: disable=missing-docstring

from nengo.exceptions import SimulationError
import numpy as np
import pytest
import tensorflow as tf

from nengo_dl import utils


def test_sanitize_name():
    assert utils.sanitize_name(0) == "0"
    assert utils.sanitize_name("a b") == "a_b"
    assert utils.sanitize_name("a:b") == "a_b"

    assert utils.sanitize_name(r"Aa0.-/\,?^&*") == r"Aa0.-/"


def test_function_names():
    def my_func(x):
        return x

    class MyFunc:
        def __call__(self, x):
            return x

    assert utils.function_name(my_func) == "my_func"
    assert utils.function_name(MyFunc()) == "MyFunc"


def test_align_func():
    def my_func():
        return [0, 1, 2, 3]

    x = utils.align_func((4,), tf.float32)(my_func)()
    assert x.shape == (4,)
    assert x.dtype == np.float32
    assert np.allclose(x, [0, 1, 2, 3])

    x = utils.align_func((2, 2), np.int64)(my_func)()
    assert x.shape == (2, 2)
    assert x.dtype == np.int64
    assert np.allclose(x, [[0, 1], [2, 3]])


@pytest.mark.parametrize("shuffle", (False, True))
def test_minibatch_generator(shuffle):
    data = {"a": np.arange(100)[:, None], "b": np.arange(1, 101)[:, None]}

    x_all = []
    y_all = []
    for _, d in utils.minibatch_generator(data, 10, shuffle=shuffle):
        x_all += [d["a"]]
        y_all += [d["b"]]

    x_all = np.concatenate(x_all)
    y_all = np.concatenate(y_all)

    if shuffle:
        assert not np.allclose(x_all, data["a"])
        assert not np.allclose(y_all, data["b"])
        x_all = np.sort(x_all, axis=0)
        y_all = np.sort(y_all, axis=0)

    assert np.allclose(x_all, data["a"])
    assert np.allclose(y_all, data["b"])

    x_all = []
    y_all = []
    with pytest.warns(UserWarning):
        for _, d in utils.minibatch_generator(data, 12, shuffle=shuffle):
            assert d["a"].shape[0] == 12
            assert d["b"].shape[0] == 12
            x_all += [d["a"]]
            y_all += [d["b"]]
    x_all = np.sort(np.concatenate(x_all))
    y_all = np.sort(np.concatenate(y_all))

    assert len(x_all) == 96
    assert len(y_all) == 96

    if shuffle:
        assert not np.allclose(x_all, np.arange(96)[:, None])
        assert not np.allclose(y_all, np.arange(1, 97)[:, None])
    else:
        assert np.allclose(x_all, np.arange(96)[:, None])
        assert np.allclose(y_all, np.arange(1, 97)[:, None])


@pytest.mark.parametrize("truncation", (None, 3, 5))
def test_truncation(truncation):
    data = {"a": np.random.randn(2, 10), "b": np.random.randn(2, 10)}

    duration = 10 if truncation is None else truncation

    with pytest.warns(None) as w:
        for i, (o, d) in enumerate(
            utils.minibatch_generator(data, 2, shuffle=False, truncation=truncation)
        ):
            assert np.allclose(d["a"], data["a"][:, o : o + duration])
            assert np.allclose(d["b"], data["b"][:, o : o + duration])

    # pylint: disable=undefined-loop-variable
    assert i == 10 // duration - (10 % duration == 0)

    assert len(w) == (1 if truncation == 3 else 0)

    # check truncation with n_steps input
    for i, (o, d) in enumerate(utils.minibatch_generator(10, None, truncation=3)):
        assert o == i * 3
        if i < 3:
            assert d == 3
        else:
            assert d == 1


def test_print_op(capsys, sess):
    x = tf.constant(0)
    y = utils.print_op(x, "hello")
    z = y + 0

    sess.run(z)

    out, _ = capsys.readouterr()

    assert out == "hello 0\n"


def test_find_non_differentiable():
    x = tf.constant(0)
    y = utils.print_op(x, "test")
    z = y + 1

    with pytest.raises(SimulationError):
        utils.find_non_differentiable([x], [z])

    x = tf.constant(0)
    y = x * 2
    z = y + 1

    utils.find_non_differentiable([x], [z])


def test_progress_bar():
    progress = utils.ProgressBar("test", max_value=10).start()

    assert progress.max_steps == progress.max_value == 10

    sub = progress.sub().start()

    # check that starting a new subprocess closes the first one
    sub2 = progress.sub().start()
    assert sub.finished

    # check that iterable wrapping works properly
    counter = 0
    for _ in progress(range(15)):
        counter += 1

    # note: progress value cut-off at 10 (the max value we specified)
    assert counter == 15
    assert progress.value == 10

    # check that closing the parent process closes the sub
    assert sub2.finished
    assert progress.finished


def test_gpu_check():
    assert utils.tf_gpu_installed == tf.test.is_gpu_available()
